# =============== Equipment Device - Mixins ===============
#
# 責任拆分 Mixin
#
# AlarmMixin: 告警管理（is_protected, active_alarms, clear_alarm, _evaluate_alarm）
# WriteMixin: 寫入管理（write, execute_action, available_actions）
#           + DO 動作控制（configure_do_actions, execute_do_action, cancel_pending_pulses）

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from csp_lib.core import get_logger
from csp_lib.equipment.alarm import AlarmEventType, AlarmState
from csp_lib.equipment.transport import WriteResult, WriteStatus

from .action import DOActionConfig, DOMode
from .events import (
    EVENT_ALARM_CLEARED,
    EVENT_ALARM_TRIGGERED,
    EVENT_WRITE_COMPLETE,
    EVENT_WRITE_ERROR,
    DeviceAlarmPayload,
    WriteCompletePayload,
    WriteErrorPayload,
)

logger = get_logger(__name__)

if TYPE_CHECKING:
    from csp_lib.equipment.alarm import AlarmEvaluator, AlarmStateManager
    from csp_lib.equipment.core import WritePoint
    from csp_lib.equipment.transport import ValidatedWriter

    from .config import DeviceConfig
    from .events import DeviceEventEmitter


class AlarmMixin:
    """
    告警管理 mixin

    提供 is_protected, active_alarms, clear_alarm, _evaluate_alarm。

    依賴：self._alarm_manager, self._alarm_evaluators, self._emitter, self._config
    """

    if TYPE_CHECKING:
        _alarm_manager: AlarmStateManager
        _config: DeviceConfig
        _emitter: DeviceEventEmitter
        _alarm_evaluators: list[AlarmEvaluator]
        _disabled_points: set[str]

    @property
    def is_protected(self) -> bool:
        """設備是否已保護"""
        return self._alarm_manager.has_protection_alarm()

    @property
    def active_alarms(self) -> list[AlarmState]:
        """啟用中的告警"""
        return self._alarm_manager.get_active_alarms()

    async def clear_alarm(self, code: str) -> None:
        """手動清除告警"""
        event = self._alarm_manager.clear_alarm(code)
        if event:
            payload = DeviceAlarmPayload(device_id=self._config.device_id, alarm_event=event)
            await self._emitter.emit_await(EVENT_ALARM_CLEARED, payload)

    async def _evaluate_alarm(self, values: dict[str, Any]) -> None:
        """評估告警（跳過 disabled 點位）"""
        for evaluator in self._alarm_evaluators:
            if evaluator.point_name in self._disabled_points:
                continue
            point_value = values.get(evaluator.point_name)
            if point_value is None:
                continue

            try:
                evaluations = evaluator.evaluate(point_value)
                events = self._alarm_manager.update(evaluations)

                for event in events:
                    payload = DeviceAlarmPayload(device_id=self._config.device_id, alarm_event=event)
                    if event.event_type == AlarmEventType.TRIGGERED:
                        await self._emitter.emit_await(EVENT_ALARM_TRIGGERED, payload)
                    else:
                        await self._emitter.emit_await(EVENT_ALARM_CLEARED, payload)
            except Exception as e:
                logger.warning(
                    f"[{self._config.device_id}] Alarm evaluation failed for point '{evaluator.point_name}': {e}"
                )


class WriteMixin:
    """
    寫入管理 mixin

    提供 write, execute_action, available_actions。

    依賴：self._write_points, self._writer, self._emitter, self._config, self.ACTIONS
    """

    if TYPE_CHECKING:
        _config: DeviceConfig
        _emitter: DeviceEventEmitter
        _write_points: dict[str, WritePoint]
        _writer: ValidatedWriter
        _disabled_points: set[str]
        _latest_values: dict[str, Any]
        ACTIONS: dict[str, str]
        _do_actions: dict[str, DOActionConfig]
        _pulse_tasks: list[asyncio.Task[None]]

    async def write(self, name: str, value: Any, verify: bool = False) -> WriteResult:
        """寫入點位值"""
        device_id = getattr(self._config, "device_id", "?")
        logger.debug(f"[{device_id}] write 開始: point={name}, value={value}, verify={verify}")

        # 檢查點位是否被停用
        if name in self._disabled_points:
            logger.warning(f"[{device_id}] write 點位已停用: {name}")
            return WriteResult(
                status=WriteStatus.VALIDATION_FAILED,
                point_name=name,
                value=value,
                error_message=f"點位 '{name}' 已停用",
            )

        point = self._write_points.get(name)
        if point is None:
            logger.warning(f"[{device_id}] write 點位不存在: {name}, 可用點位: {list(self._write_points.keys())}")
            return WriteResult(
                status=WriteStatus.VALIDATION_FAILED,
                point_name=name,
                value=value,
                error_message=f"寫入點位 {name} 失敗，點位不存在",
            )

        logger.debug(f"[{device_id}] write 點位已找到: {name}, address={point.address}, fc={point.function_code}")
        result = await self._writer.write(point=point, value=value, verify=verify)
        logger.debug(
            f"[{device_id}] write 結果: point={name}, status={result.status.value}, error={result.error_message or 'none'}"
        )

        if result.status == WriteStatus.SUCCESS:
            self._emitter.emit(
                EVENT_WRITE_COMPLETE,
                WriteCompletePayload(device_id=self._config.device_id, point_name=name, value=value),
            )
        else:
            self._emitter.emit(
                EVENT_WRITE_ERROR,
                WriteErrorPayload(
                    device_id=self._config.device_id, point_name=name, value=value, error=result.error_message
                ),
            )

        return result

    async def execute_action(self, action: str, **params: Any) -> WriteResult:
        """
        執行高階動作

        根據 ACTIONS 映射呼叫對應的方法。

        Args:
            action: 動作名稱（如 "start", "stop"）
            **params: 傳遞給方法的參數

        Returns:
            WriteResult：成功時 status=SUCCESS，失敗時包含錯誤訊息

        Raises:
            不拋出異常，所有錯誤透過 WriteResult 回傳
        """
        device_id = getattr(self._config, "device_id", "?")
        logger.debug(f"[{device_id}] execute_action 開始: action={action}, params={params}")

        method_name = self.ACTIONS.get(action)
        if method_name is None:
            logger.warning(f"[{device_id}] action 不支援: {action}, 可用: {list(self.ACTIONS.keys())}")
            return WriteResult(
                status=WriteStatus.VALIDATION_FAILED,
                point_name=action,
                value=None,
                error_message=f"Action '{action}' not supported. Available: {list(self.ACTIONS.keys())}",
            )

        method = getattr(self, method_name, None)
        if method is None or not callable(method):
            logger.error(f"[{device_id}] action 方法不存在: action={action} → method={method_name}")
            return WriteResult(
                status=WriteStatus.VALIDATION_FAILED,
                point_name=action,
                value=None,
                error_message=f"Method '{method_name}' not found for action '{action}'",
            )

        try:
            logger.debug(f"[{device_id}] 呼叫方法: {method_name}(**{params})")
            result = await method(**params)

            # 若方法回傳 WriteResult，直接使用其狀態
            if isinstance(result, WriteResult):
                if result.status != WriteStatus.SUCCESS:
                    logger.error(
                        f"[{device_id}] execute_action 失敗: action={action}, status={result.status.value}, error={result.error_message}"
                    )
                else:
                    logger.debug(f"[{device_id}] execute_action 成功: action={action}")
                return result

            logger.debug(f"[{device_id}] execute_action 成功: action={action}")
            return WriteResult(
                status=WriteStatus.SUCCESS,
                point_name=action,
                value=params if params else None,
            )
        except Exception as e:
            logger.error(f"[{device_id}] execute_action 失敗: action={action}, error={e}")
            return WriteResult(
                status=WriteStatus.WRITE_FAILED,
                point_name=action,
                value=params if params else None,
                error_message=str(e),
            )

    @property
    def available_actions(self) -> list[str]:
        """取得支援的動作列表"""
        return list(self.ACTIONS.keys())

    # =============== DO Action ===============

    def configure_do_actions(self, configs: Sequence[DOActionConfig]) -> None:
        """配置 DO 動作

        Args:
            configs: DO 動作配置列表

        Raises:
            ValueError: 有重複的 label
        """
        labels = [c.label for c in configs]
        if len(labels) != len(set(labels)):
            raise ValueError(f"Duplicate DO action labels: {labels}")
        for task in self._pulse_tasks:
            task.cancel()
        self._do_actions = {c.label: c for c in configs}
        self._pulse_tasks = []

    @property
    def available_do_actions(self) -> list[DOActionConfig]:
        """可用的 DO 動作列表"""
        return list(self._do_actions.values())

    async def execute_do_action(self, label: str, *, turn_off: bool = False) -> WriteResult:
        """執行指定 DO 動作

        Args:
            label: 動作標籤
            turn_off: 是否執行關閉動作（寫入 off_value）。
                SUSTAINED: turn_off=False 寫 on_value，turn_off=True 寫 off_value。
                TOGGLE: 忽略此參數，自動反轉。
                PULSE: 忽略此參數，固定 on → 延遲 → off。

        Returns:
            WriteResult

        Raises:
            ValueError: label 不存在
        """
        config = self._do_actions.get(label)
        if config is None:
            raise ValueError(f"Unknown DO action: {label!r}. Available: {list(self._do_actions.keys())}")

        if config.mode == DOMode.SUSTAINED:
            value = config.off_value if turn_off else config.on_value
            return await self.write(config.point_name, value)

        if config.mode == DOMode.TOGGLE:
            current = self._latest_values.get(config.point_name, config.off_value)
            new_value = config.off_value if current == config.on_value else config.on_value
            result = await self.write(config.point_name, new_value)
            if result.status == WriteStatus.SUCCESS:
                self._latest_values[config.point_name] = new_value
            return result

        # DOMode.PULSE
        result = await self.write(config.point_name, config.on_value)
        if result.status != WriteStatus.SUCCESS:
            return result
        task = asyncio.create_task(self._pulse_off(config))
        self._pulse_tasks.append(task)
        task.add_done_callback(lambda t: self._pulse_tasks.remove(t) if t in self._pulse_tasks else None)
        return result

    async def _pulse_off(self, config: DOActionConfig) -> None:
        """PULSE 模式：等待 pulse_duration 後寫回 off_value"""
        try:
            await asyncio.sleep(config.pulse_duration)
            await self.write(config.point_name, config.off_value)
        except asyncio.CancelledError:
            pass
        except Exception:
            logger.opt(exception=True).warning(f"Pulse off failed for {config.label}")

    async def cancel_pending_pulses(self) -> None:
        """取消所有進行中的 PULSE 任務（shutdown 時呼叫）"""
        for task in list(self._pulse_tasks):
            task.cancel()
        if self._pulse_tasks:
            await asyncio.gather(*self._pulse_tasks, return_exceptions=True)
        self._pulse_tasks = []


__all__ = [
    "AlarmMixin",
    "WriteMixin",
]



# =============== Ramp Stop Strategy ===============
#
# 斜坡降功率策略（維護型）
#
# 被 EventDrivenOverride push 後，從 last_command 開始按斜率降至 0。
# 使用實際 dt（monotonic clock）計算每步降幅。
#
# 與 StopStrategy 的區別：
#   StopStrategy: 立即 P=0（用於設備告警等需要立即停止的情境）
#   RampStopStrategy: 斜坡降至 0（用於通訊中斷等可容忍漸進停止的情境）
#
# 搭配 EventDrivenOverride + ModeManager 使用：
#   controller.register_mode("ramp_stop", RampStopStrategy(2000), ModePriority.PROTECTION)
#   controller.register_event_override(ContextKeyOverride(name="ramp_stop", key="comm_timeout"))

from __future__ import annotations

import time

from csp_lib.controller.core import Command, ExecutionConfig, ExecutionMode, Strategy, StrategyContext
from csp_lib.core import get_logger

logger = get_logger(__name__)


class RampStopStrategy(Strategy):
    """
    斜坡降功率策略

    透過 ModeManager.push_override() 啟動，從當前功率開始按斜率降至 0。
    到達 0 後維持 P=0 直到被 pop_override() 移除。

    使用實際 dt（monotonic clock）計算每步降幅，不依賴固定 interval：
        ramp_step = ramp_rate_pct / 100 × rated_power × dt

    Args:
        rated_power: 系統額定功率 (kW)
        ramp_rate_pct: 斜率 (%/s)，預設 5.0 表示每秒降 5% 額定功率
    """

    def __init__(self, rated_power: float, ramp_rate_pct: float = 5.0) -> None:
        self._rated = rated_power
        self._ramp_rate = ramp_rate_pct
        self._current_p: float = 0.0
        self._last_time: float | None = None

    @property
    def execution_config(self) -> ExecutionConfig:
        return ExecutionConfig(mode=ExecutionMode.PERIODIC, interval_seconds=1)

    def execute(self, context: StrategyContext) -> Command:
        now = time.monotonic()
        dt = (now - self._last_time) if self._last_time is not None else 0.0
        self._last_time = now

        # 第一次執行：繼承上次功率
        if self._current_p == 0.0 and dt == 0.0:
            self._current_p = context.last_command.p_target
            logger.info(f"RampStop: starting ramp from {self._current_p:.1f}kW")

        # 已到 0 → 維持
        if self._current_p == 0.0:
            return Command(p_target=0.0, q_target=0.0)

        # 計算 ramp step
        ramp_step = self._ramp_rate / 100.0 * self._rated * dt

        if abs(self._current_p) <= ramp_step:
            self._current_p = 0.0
            logger.info("RampStop: reached P=0")
        elif self._current_p > 0:
            self._current_p -= ramp_step
        else:
            self._current_p += ramp_step

        logger.debug(f"RampStop: P={self._current_p:.1f}kW (dt={dt:.3f}s, step={ramp_step:.1f}kW)")
        return Command(p_target=self._current_p, q_target=0.0)

    async def on_activate(self) -> None:
        self._current_p = 0.0
        self._last_time = None
        logger.info(f"RampStopStrategy activated (rated={self._rated}kW, rate={self._ramp_rate}%/s)")

    async def on_deactivate(self) -> None:
        self._current_p = 0.0
        self._last_time = None
        logger.info("RampStopStrategy deactivated")

    def __str__(self) -> str:
        return f"RampStopStrategy(rated={self._rated}kW, rate={self._ramp_rate}%/s)"


__all__ = ["RampStopStrategy"]



# =============== Manager Command - Adapters ===============
#
# 指令適配器子模組

from .config import CommandAdapterConfig
from .redis import CommandResult, RedisCommandAdapter

__all__ = [
    "CommandAdapterConfig",
    "CommandResult",
    "RedisCommandAdapter",
]



"""WebSocket endpoint."""

from __future__ import annotations

from fastapi import APIRouter, WebSocket, WebSocketDisconnect

from ..dependencies import get_ws_manager

router = APIRouter()


@router.websocket("/ws")
async def websocket_endpoint(ws: WebSocket) -> None:
    """WebSocket endpoint for real-time event streaming."""
    manager = get_ws_manager(ws)  # type: ignore[arg-type]  # WebSocket has same .app.state as Request
    await manager.connect(ws)
    try:
        while True:
            # Keep alive - receive client pings/messages
            await ws.receive_text()
    except WebSocketDisconnect:
        pass
    finally:
        await manager.disconnect(ws)



"""
Built-in DataSyncSource implementations.

DataSyncSource 將外部資料來源（Redis / 輪詢回呼）的數據同步到 gateway 的 Input Registers。
"""

from __future__ import annotations

import asyncio
import json
from collections.abc import Awaitable, Callable
from typing import Any

from csp_lib.core import get_logger

logger = get_logger(__name__)

# Type alias matching protocol.py
UpdateRegisterCallback = Callable[[str, Any], Awaitable[None]]


class RedisSubscriptionSource:
    """
    Subscribes to a Redis channel and updates registers from published messages.

    Expected message format (JSON)::

        {"register": "<name>", "value": <physical_value>}

    Or batch::

        [{"register": "<name>", "value": <physical_value>}, ...]

    Or key-value shorthand::

        {"<register_name>": <physical_value>, ...}

    Satisfies DataSyncSource protocol.

    Args:
        redis_client: redis.asyncio.Redis 或 csp_lib.redis.RedisClient 實例
        channel: Redis channel 名稱
        batch_mode: 若 True，將 payload 視為 {name: value} dict（簡潔模式）
    """

    def __init__(self, redis_client: Any, channel: str = "gateway:sync", batch_mode: bool = True) -> None:
        self._redis = redis_client
        self._channel = channel
        self._batch_mode = batch_mode
        self._update_cb: UpdateRegisterCallback | None = None
        self._task: asyncio.Task | None = None

    async def start(self, update_callback: UpdateRegisterCallback) -> None:
        self._update_cb = update_callback
        self._task = asyncio.create_task(self._listen_loop(), name=f"redis_sync_{self._channel}")
        logger.info(f"RedisSubscriptionSource started: channel={self._channel}")

    async def stop(self) -> None:
        if self._task and not self._task.done():
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
        logger.info(f"RedisSubscriptionSource stopped: channel={self._channel}")

    async def _listen_loop(self) -> None:
        # Get pubsub - support both RedisClient and raw redis.asyncio.Redis
        if hasattr(self._redis, "pubsub"):
            pubsub = self._redis.pubsub()
        else:
            pubsub = self._redis.pubsub()

        await pubsub.subscribe(self._channel)
        try:
            while True:
                msg = await pubsub.get_message(ignore_subscribe_messages=True, timeout=1.0)
                if msg and msg["type"] == "message":
                    await self._handle_message(msg["data"])
        except asyncio.CancelledError:
            pass
        except Exception:
            logger.opt(exception=True).error(f"RedisSubscriptionSource listen error: {self._channel}")
        finally:
            try:
                await pubsub.unsubscribe(self._channel)
                await pubsub.aclose()
            except Exception:
                pass

    async def _handle_message(self, data: Any) -> None:
        if self._update_cb is None:
            return
        try:
            if isinstance(data, bytes):
                data = data.decode()
            payload = json.loads(data)

            if isinstance(payload, dict) and "register" in payload and "value" in payload:
                # Single: {"register": "soc", "value": 75.5}
                # 優先匹配，避免被 batch_mode shorthand 吞掉
                name = payload["register"]
                value = payload["value"]
                try:
                    await self._update_cb(name, value)
                except KeyError:
                    logger.debug(f"RedisSync: unknown register '{name}', skipping")
            elif isinstance(payload, list):
                # Batch: [{"register": "soc", "value": 75.5}, ...]
                for item in payload:
                    name = item.get("register")
                    value = item.get("value")
                    if name is not None and value is not None:
                        try:
                            await self._update_cb(name, value)
                        except KeyError:
                            logger.debug(f"RedisSync: unknown register '{name}', skipping")
            elif self._batch_mode and isinstance(payload, dict):
                # Key-value shorthand: {"soc": 75.5, "soh": 98.0}
                for name, value in payload.items():
                    try:
                        await self._update_cb(name, value)
                    except KeyError:
                        logger.debug(f"RedisSync: unknown register '{name}', skipping")

        except (json.JSONDecodeError, TypeError, KeyError):
            logger.opt(exception=True).warning(f"RedisSync: invalid message on {self._channel}")


class PollingCallbackSource:
    """
    Periodically calls a user-provided async function to get register values.

    The callback returns a dict of {register_name: physical_value}.
    Only non-None values are written.

    Satisfies DataSyncSource protocol.

    Args:
        callback: async function() -> dict[str, Any]
        interval: polling interval in seconds
    """

    def __init__(
        self,
        callback: Callable[[], Awaitable[dict[str, Any]]],
        interval: float = 1.0,
    ) -> None:
        self._callback = callback
        self._interval = interval
        self._update_cb: UpdateRegisterCallback | None = None
        self._task: asyncio.Task | None = None

    async def start(self, update_callback: UpdateRegisterCallback) -> None:
        self._update_cb = update_callback
        self._task = asyncio.create_task(self._poll_loop(), name="polling_sync")
        logger.info(f"PollingCallbackSource started: interval={self._interval}s")

    async def stop(self) -> None:
        if self._task and not self._task.done():
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
        logger.info("PollingCallbackSource stopped")

    async def _poll_loop(self) -> None:
        try:
            while True:
                try:
                    values = await self._callback()
                    if self._update_cb and values:
                        for name, value in values.items():
                            if value is not None:
                                try:
                                    await self._update_cb(name, value)
                                except KeyError:
                                    logger.debug(f"PollingSync: unknown register '{name}', skipping")
                except Exception:
                    logger.opt(exception=True).warning("PollingCallbackSource: callback error")
                await asyncio.sleep(self._interval)
        except asyncio.CancelledError:
            pass


__all__ = [
    "RedisSubscriptionSource",
    "PollingCallbackSource",
]



# =============== Island Mode Strategy ===============
#
# 離網模式策略 (Grid Forming)
# - 啟用時: 切離 ACB (set_open)
# - 停用時: 等待 sync_ok 後搭接 ACB (set_close)
#
# 使用 RelayProtocol 避免直接依賴 P3U30

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Optional, Protocol, runtime_checkable

from csp_lib.controller.core import (
    Command,
    ConfigMixin,
    ExecutionConfig,
    ExecutionMode,
    Strategy,
    StrategyContext,
)
from csp_lib.core import get_logger

logger = get_logger(__name__)


# =============== Relay Protocol ===============


@runtime_checkable
class RelayProtocol(Protocol):
    """
    繼電器/斷路器控制協定

    用於 IslandModeStrategy 的 ACB 控制。
    實作此協定的設備需提供以下屬性和方法：
    - sync_ok: 同步狀態
    - sync_counter: 同步計數
    - set_open(): 開啟斷路器
    - set_close(): 閉合斷路器 (需 sync_ok)
    - set_force_close(): 強制閉合斷路器
    """

    @property
    def sync_ok(self) -> bool:
        """同步狀態"""
        ...

    @property
    def sync_counter(self) -> int:
        """同步計數"""
        ...

    async def set_open(self) -> None:
        """開啟斷路器"""
        ...

    async def set_close(self) -> None:
        """閉合斷路器 (需 sync_ok)"""
        ...

    async def set_force_close(self) -> None:
        """強制閉合斷路器"""
        ...


# =============== Island Mode Config ===============


@dataclass
class IslandModeConfig(ConfigMixin):
    """
    離網模式配置

    Attributes:
        sync_timeout: 等待 sync_ok 超時 (秒)，預設 60
    """

    sync_timeout: float = 60.0


# =============== Island Mode Strategy ===============


class IslandModeStrategy(Strategy):
    """
    離網模式策略 (Grid Forming / Island Mode)

    當策略啟用時，自動切離 ACB 進入離網模式。
    當策略停用時，等待 sync_ok 後自動搭接 ACB 返回併網模式。

    **重要**: 此策略使用 TRIGGERED 模式，不會主動發送命令。
    功率控制應由 PCS 自身的 VF 模式處理，或由其他策略組合使用。

    Usage:
        relay = P3U30(...)  # 實作 RelayProtocol
        strategy = IslandModeStrategy(relay, config=IslandModeConfig())

        # 啟用 -> 自動切離 ACB
        await executor.set_strategy(strategy)

        # 切換策略 -> 自動等待 sync_ok 後搭接 ACB
        await executor.set_strategy(other_strategy)
    """

    def __init__(
        self,
        relay: RelayProtocol,
        config: Optional[IslandModeConfig] = None,
    ) -> None:
        """
        初始化離網模式策略

        Args:
            relay: 實作 RelayProtocol 的設備 (如 P3U30)
            config: 離網模式配置
        """
        self._relay = relay
        self._config = config or IslandModeConfig()

    @property
    def config(self) -> IslandModeConfig:
        """當前配置"""
        return self._config

    @property
    def execution_config(self) -> ExecutionConfig:
        """執行配置: TRIGGERED 模式 - 不主動執行"""
        return ExecutionConfig(mode=ExecutionMode.TRIGGERED, interval_seconds=1)

    def execute(self, context: StrategyContext) -> Command:
        """
        執行策略邏輯

        離網模式不發送功率命令，返回 last_command 維持現狀。
        """
        return context.last_command

    async def on_activate(self) -> None:
        """
        策略啟用: 切離 ACB

        開啟 ACB，使系統進入離網 (孤島) 模式。
        """
        logger.info("IslandModeStrategy: 啟用離網模式，切離 ACB")
        await self._relay.set_open()
        logger.info("IslandModeStrategy: ACB 已開啟，進入離網模式")

    async def on_deactivate(self) -> None:
        """
        策略停用: 等待 sync_ok 後搭接 ACB

        持續等待同步信號，確認後閉合 ACB 返回併網模式。
        """
        logger.info("IslandModeStrategy: 準備離開離網模式，等待 sync_ok...")

        # 等待 sync_ok
        timeout = self._config.sync_timeout
        elapsed = 0.0
        check_interval = 0.5

        while elapsed < timeout:
            if self._relay.sync_ok:
                break
            logger.debug(
                f"IslandModeStrategy: 等待同步中... counter={self._relay.sync_counter}, elapsed={elapsed:.1f}s"
            )
            await asyncio.sleep(check_interval)
            elapsed += check_interval

        if not self._relay.sync_ok:
            logger.critical(f"IslandModeStrategy: 等待 sync_ok 超時 ({timeout}s)，失敗，請手動處理 ACB")
        else:
            logger.info("IslandModeStrategy: sync_ok 確認，閉合 ACB")
            await self._relay.set_close()
            logger.info("IslandModeStrategy: ACB 已閉合，返回併網模式")

    def update_config(self, config: IslandModeConfig) -> None:
        """更新配置"""
        self._config = config

    def __str__(self) -> str:
        return "IslandModeStrategy()"



# =============== Strategy Module ===============
#
# 策略抽象基礎類別
# - Strategy: 策略抽象基礎類別

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from .command import Command
from .context import StrategyContext
from .execution import ExecutionConfig

if TYPE_CHECKING:
    from csp_lib.equipment.device.capability import Capability


class Strategy(ABC):
    """
    策略抽象基礎類別

    所有策略必須繼承此類別並實作：
    - execution_config: 回傳執行配置
    - execute(): 執行策略邏輯並回傳 Command

    可選覆寫：
    - on_activate(): 策略啟用時呼叫
    - on_deactivate(): 策略停用時呼叫

    擴充性設計：
    - 策略可在 execute() 中呼叫其他策略的 execute() 取得 Command
    - 自行組合多個 Command 達成複合策略效果
    """

    @property
    @abstractmethod
    def execution_config(self) -> ExecutionConfig:
        """回傳策略的執行配置"""
        pass

    @abstractmethod
    def execute(self, context: StrategyContext) -> Command:
        """
        執行策略邏輯

        Args:
            context: 執行時上下文，包含 last_command、soc 等狀態

        Returns:
            Command: 策略輸出命令
        """
        pass

    @property
    def required_capabilities(self) -> tuple[Capability, ...]:
        """
        此策略需要的能力。
        SystemController 在 register_mode() 時驗證。
        預設空 tuple，不影響既有策略。
        """
        return ()

    @property
    def suppress_heartbeat(self) -> bool:
        """
        是否在此策略啟用期間暫停心跳寫入

        預設 False。旁路模式等需要完全停止控制器輸出的策略應覆寫為 True。
        """
        return False

    async def on_activate(self) -> None:
        """
        策略啟用時呼叫 (可選覆寫)

        用途：初始化內部狀態、執行啟動動作等
        """
        pass

    async def on_deactivate(self) -> None:
        """
        策略停用時呼叫 (可選覆寫)

        用途：清理資源、執行收尾動作等
        """
        pass

    def __str__(self) -> str:
        return f"{self.__class__.__name__}"

    def __repr__(self) -> str:
        return self.__str__()



# =============== Modbus Data Types - Dynamic ===============
#
# 動態長度整數類型實作
#
# 支援任意 16 的倍數位元寬度：
#   - DynamicInt: 動態長度有號整數
#   - DynamicUInt: 動態長度無號整數
#
# 使用範例：
#   uint48 = DynamicUInt(48)  # 48-bit 無號整數，需要 3 個暫存器

from __future__ import annotations

from ..enums import ByteOrder, RegisterOrder
from ..exceptions import ModbusConfigError, ModbusDecodeError, ModbusEncodeError
from .base import ModbusDataType


class _DynamicIntBase(ModbusDataType):
    """
    動態長度整數內部基類

    以 signed 參數化，統一有號/無號整數的暫存器拆分/組合邏輯。
    """

    def __init__(self, bit_width: int, *, signed: bool) -> None:
        if bit_width <= 0:
            raise ModbusConfigError(f"bit_width 必須為正整數，收到: {bit_width}")
        if bit_width % 16 != 0:
            raise ModbusConfigError(f"bit_width 必須為 16 的倍數，收到: {bit_width}")

        self._bit_width = bit_width
        self._register_count = bit_width // 16
        self._signed = signed

        if signed:
            self._max_value = (1 << (bit_width - 1)) - 1
            self._min_value = -(1 << (bit_width - 1))
        else:
            self._max_value = (1 << bit_width) - 1
            self._min_value = 0

    @property
    def register_count(self) -> int:
        return self._register_count

    @property
    def bit_width(self) -> int:
        """位元寬度"""
        return self._bit_width

    def _type_label(self) -> str:
        name = "DynamicInt" if self._signed else "DynamicUInt"
        return f"{name}({self._bit_width})"

    def encode(
        self,
        value: int,
        byte_order: ByteOrder,
        register_order: RegisterOrder,
    ) -> list[int]:
        if not isinstance(value, int):
            raise ModbusEncodeError(f"{self._type_label()} 需要整數，收到: {type(value).__name__}")
        if not self._min_value <= value <= self._max_value:
            raise ModbusEncodeError(f"{self._type_label()} 範圍為 {self._min_value}~{self._max_value}，收到: {value}")

        # 處理負數：轉換為補數表示
        raw = value
        if self._signed and raw < 0:
            raw = (1 << self._bit_width) + raw

        # 將整數分割為多個 16-bit 暫存器 (LSW first)
        registers = []
        for _ in range(self._register_count):
            registers.append(raw & 0xFFFF)
            raw >>= 16

        # 預設 LSW first，若 HIGH_FIRST 則反轉為 MSW first
        if register_order == RegisterOrder.HIGH_FIRST:
            registers.reverse()

        return registers

    def decode(
        self,
        registers: list[int],
        byte_order: ByteOrder,
        register_order: RegisterOrder,
    ) -> int:
        if len(registers) < self._register_count:
            raise ModbusDecodeError(
                f"{self._type_label()} 需要 {self._register_count} 個暫存器，收到: {len(registers)}"
            )

        regs = list(registers[: self._register_count])

        # 還原順序：若 HIGH_FIRST 則反轉回 LSW first 以便組合
        if register_order == RegisterOrder.HIGH_FIRST:
            regs.reverse()

        # 組合為整數 (從 LSW 開始)
        value = 0
        for i, reg in enumerate(regs):
            value |= reg << (16 * i)

        # 處理負數：從補數還原
        if self._signed and value >= (1 << (self._bit_width - 1)):
            value -= 1 << self._bit_width

        return value


class DynamicInt(_DynamicIntBase):
    """
    動態長度有號整數

    支援任意 16 的倍數位元寬度。

    Args:
        bit_width: 位元寬度，必須為 16 的倍數

    Raises:
        ModbusConfigError: bit_width 非 16 的倍數

    使用範例：
        >>> int48 = DynamicInt(48)
        >>> int48.register_count
        3
    """

    def __init__(self, bit_width: int) -> None:
        super().__init__(bit_width, signed=True)


class DynamicUInt(_DynamicIntBase):
    """
    動態長度無號整數

    支援任意 16 的倍數位元寬度。

    Args:
        bit_width: 位元寬度，必須為 16 的倍數

    Raises:
        ModbusConfigError: bit_width 非 16 的倍數

    使用範例：
        >>> uint48 = DynamicUInt(48)
        >>> uint48.register_count
        3
    """

    def __init__(self, bit_width: int) -> None:
        super().__init__(bit_width, signed=False)


__all__ = [
    "DynamicInt",
    "DynamicUInt",
]



# =============== Manager Alarm - Config ===============
#
# 告警持久化配置
#
# 定義告警持久化相關參數：
#   - AlarmPersistenceConfig: 斷線告警代碼與名稱

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class AlarmPersistenceConfig:
    """
    告警持久化配置

    Attributes:
        disconnect_code: 斷線告警的固定代碼
        disconnect_name: 斷線告警的顯示名稱
    """

    disconnect_code: str = "DISCONNECT"
    disconnect_name: str = "設備斷線"

    def __post_init__(self) -> None:
        if not self.disconnect_code:
            raise ValueError("disconnect_code 不可為空")
        if not self.disconnect_name:
            raise ValueError("disconnect_name 不可為空")


__all__ = [
    "AlarmPersistenceConfig",
]



# =============== Manager - Unified ===============
#
# 統一設備管理器
#
# 整合所有子管理器，提供單一入口點：
#   - UnifiedConfig: 統一管理器配置
#   - UnifiedDeviceManager: 統一設備管理器

from __future__ import annotations

import threading
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Sequence

from csp_lib.core import AsyncLifecycleMixin, get_logger

from .alarm import AlarmPersistenceManager, AlarmRepository
from .command import CommandRepository, WriteCommandManager
from .data import DataUploadManager
from .device import DeviceManager
from .state import StateSyncManager

if TYPE_CHECKING:
    from csp_lib.equipment.device import AsyncModbusDevice
    from csp_lib.integration.registry import DeviceRegistry
    from csp_lib.manager.base import BatchUploader
    from csp_lib.notification import NotificationDispatcher
    from csp_lib.redis import RedisClient
    from csp_lib.statistics import StatisticsConfig, StatisticsManager

logger = get_logger(__name__)


# ================ 配置 ================


@dataclass
class UnifiedConfig:
    """
    統一管理器配置

    所有子管理器皆為可選，未配置的功能將自動跳過。

    Attributes:
        alarm_repository: 告警持久化 Repository（可選）
        command_repository: 寫入指令 Repository（可選）
        mongo_uploader: MongoDB 批次上傳器（可選）
        redis_client: Redis 客戶端（可選）
        notification_dispatcher: 通知分發器（可選）

    Example:
        ```python
        config = UnifiedConfig(
            alarm_repository=mongo_alarm_repo,
            command_repository=mongo_cmd_repo,
            mongo_uploader=uploader,
            redis_client=redis,
        )
        ```
    """

    alarm_repository: AlarmRepository | None = None
    command_repository: CommandRepository | None = None
    mongo_uploader: BatchUploader | None = None
    redis_client: RedisClient | None = None
    notification_dispatcher: NotificationDispatcher | None = None
    statistics_config: StatisticsConfig | None = None
    device_registry: DeviceRegistry | None = None


# ================ 統一管理器 ================


class UnifiedDeviceManager(AsyncLifecycleMixin):
    """
    統一設備管理器

    整合 DeviceManager、AlarmPersistenceManager、WriteCommandManager、
    DataUploadManager、StateSyncManager、StatisticsManager，提供單一入口點。

    註冊設備後自動串接所有已啟用的功能，使用者無需手動 subscribe 各子管理器。

    Attributes:
        device_manager: 設備讀取管理器
        alarm_manager: 告警持久化管理器（可能為 None）
        command_manager: 寫入指令管理器（可能為 None）
        data_manager: 資料上傳管理器（可能為 None）
        state_manager: 狀態同步管理器（可能為 None）
        statistics_manager: 統計管理器（可能為 None）

    Example:
        ```python
        config = UnifiedConfig(
            alarm_repository=mongo_alarm_repo,
            command_repository=mongo_cmd_repo,
            mongo_uploader=uploader,
            redis_client=redis,
        )

        manager = UnifiedDeviceManager(config)

        # 註冊時指定 collection_name
        manager.register(meter_device, collection_name="meter")
        manager.register(io_device, collection_name="io")

        # 群組註冊
        manager.register_group([rtu1, rtu2], collection_name="rtu_data")

        async with manager:
            await asyncio.sleep(3600)
        ```
    """

    def __init__(self, config: UnifiedConfig) -> None:
        """
        初始化統一設備管理器

        Args:
            config: 統一管理器配置
        """
        self._config = config
        self._device_manager = DeviceManager()

        # 根據配置初始化子管理器（可選）
        self._alarm_manager: AlarmPersistenceManager | None = (
            AlarmPersistenceManager(config.alarm_repository, config.notification_dispatcher)
            if config.alarm_repository
            else None
        )
        self._command_manager: WriteCommandManager | None = (
            WriteCommandManager(config.command_repository) if config.command_repository else None
        )
        self._data_manager: DataUploadManager | None = (
            DataUploadManager(config.mongo_uploader) if config.mongo_uploader else None
        )
        self._state_manager: StateSyncManager | None = (
            StateSyncManager(config.redis_client) if config.redis_client else None
        )

        # Statistics manager（需要 mongo_uploader + statistics_config + csp_lib.statistics）
        self._statistics_manager: StatisticsManager | None = None
        if config.statistics_config and config.mongo_uploader:
            try:
                from csp_lib.statistics import StatisticsManager

                self._statistics_manager = StatisticsManager(
                    config.statistics_config, config.mongo_uploader, config.device_registry
                )
            except ImportError:
                logger.warning("Statistics module not available, skipping StatisticsManager initialization")

        self._register_lock = threading.Lock()

        logger.info(
            f"UnifiedDeviceManager 初始化: "
            f"alarm={self._alarm_manager is not None}, "
            f"command={self._command_manager is not None}, "
            f"data={self._data_manager is not None}, "
            f"state={self._state_manager is not None}, "
            f"statistics={self._statistics_manager is not None}"
        )

    # ================ 註冊 ================

    def register(
        self,
        device: AsyncModbusDevice,
        collection_name: str | None = None,
        traits: Sequence[str] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """
        註冊獨立設備

        設備將使用自己的 read_loop 進行讀取，並自動訂閱所有已啟用的子管理器。
        若配置了 DeviceRegistry，會同時將設備註冊到 Registry。

        Args:
            device: Modbus 設備
            collection_name: MongoDB collection 名稱（Data Upload 用，選填）
            traits: 設備 trait 標籤列表（選填，用於 DeviceRegistry）
            metadata: 設備靜態資訊（選填，用於 DeviceRegistry）
        """
        with self._register_lock:
            self._device_manager.register(device)
            self._subscribe_all(device, collection_name)
            self._register_to_registry(device, traits, metadata)
        logger.info(f"UnifiedDeviceManager: 已註冊設備 {device.device_id}")

    def register_group(
        self,
        devices: Sequence[AsyncModbusDevice],
        interval: float = 1.0,
        collection_name: str | None = None,
        traits: Sequence[str] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """
        註冊設備群組

        群組內設備將順序讀取，並自動訂閱所有已啟用的子管理器。
        群組內設備共用同一 collection_name。
        若配置了 DeviceRegistry，所有設備會同時註冊（共用相同 traits/metadata）。

        Args:
            devices: 設備列表（必須共用同一 Client）
            interval: 完整讀取一輪的間隔時間（秒）
            collection_name: MongoDB collection 名稱（群組共用，選填）
            traits: 設備 trait 標籤列表（選填，套用到群組所有設備）
            metadata: 設備靜態資訊（選填，套用到群組所有設備）
        """
        with self._register_lock:
            self._device_manager.register_group(devices, interval)
            for device in devices:
                self._subscribe_all(device, collection_name)
                self._register_to_registry(device, traits, metadata)
        device_ids = [d.device_id for d in devices]
        logger.info(f"UnifiedDeviceManager: 已註冊設備群組 {device_ids}")

    def _register_to_registry(
        self,
        device: AsyncModbusDevice,
        traits: Sequence[str] | None,
        metadata: dict[str, Any] | None,
    ) -> None:
        """
        若配置了 DeviceRegistry，將設備註冊到 Registry。

        Args:
            device: Modbus 設備
            traits: trait 標籤列表（可選）
            metadata: 靜態資訊（可選）
        """
        if self._config.device_registry is not None:
            self._config.device_registry.register(
                device,
                traits=list(traits) if traits else [],
                metadata=metadata or {},
            )

    def _subscribe_all(
        self,
        device: AsyncModbusDevice,
        collection_name: str | None,
    ) -> None:
        """
        訂閱所有已啟用的子管理器

        Args:
            device: Modbus 設備
            collection_name: MongoDB collection 名稱（可選）
        """
        if self._alarm_manager:
            self._alarm_manager.subscribe(device)

        if self._command_manager:
            self._command_manager.register_device(device)

        if self._data_manager and collection_name:
            self._data_manager.configure(device.device_id, collection_name)
            self._data_manager.subscribe(device)

        if self._state_manager:
            self._state_manager.subscribe(device)

        if self._statistics_manager:
            self._statistics_manager.subscribe(device)

    # ================ 生命週期 ================

    async def _on_start(self) -> None:
        """
        啟動管理器

        啟動所有已註冊設備的讀取循環。
        """
        await self._device_manager.start()
        logger.info("UnifiedDeviceManager 已啟動")

    async def _on_stop(self) -> None:
        """
        停止管理器

        停止所有讀取循環並斷開連線。
        """
        await self._device_manager.stop()
        logger.info("UnifiedDeviceManager 已停止")

    # ================ 屬性 ================

    @property
    def device_manager(self) -> DeviceManager:
        """設備讀取管理器"""
        return self._device_manager

    @property
    def alarm_manager(self) -> AlarmPersistenceManager | None:
        """告警持久化管理器（可能為 None）"""
        return self._alarm_manager

    @property
    def command_manager(self) -> WriteCommandManager | None:
        """寫入指令管理器（可能為 None）"""
        return self._command_manager

    @property
    def data_manager(self) -> DataUploadManager | None:
        """資料上傳管理器（可能為 None）"""
        return self._data_manager

    @property
    def state_manager(self) -> StateSyncManager | None:
        """狀態同步管理器（可能為 None）"""
        return self._state_manager

    @property
    def statistics_manager(self) -> StatisticsManager | None:
        """統計管理器（可能為 None）"""
        return self._statistics_manager

    @property
    def is_running(self) -> bool:
        """管理器是否運行中"""
        return self._device_manager.is_running

    def __repr__(self) -> str:
        return (
            f"<UnifiedDeviceManager "
            f"devices={self._device_manager.standalone_count + sum(len(g.devices) for g in self._device_manager.groups)} "
            f"running={self.is_running}>"
        )


__all__ = [
    "UnifiedConfig",
    "UnifiedDeviceManager",
]



# =============== Manager Device - Group ===============
#
# 設備群組（純順序讀取器）
#
# 用於管理需要順序讀取的設備：
#   - DeviceGroup: 設備群組類別
#
# 使用場景：
#   - RTU 通訊：多設備共用 RS485 線路
#   - Shared TCP：多設備共用 TCP Client（如 Gateway）
#   - 任何需要順序讀取的場景
#
# 設計理念：
#   - DeviceGroup 只負責「順序讀取」
#   - 連線/斷線/重連由各 Device 自己管理
#   - 不限制設備是否共用 Client

from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from csp_lib.core import get_logger

if TYPE_CHECKING:
    from csp_lib.equipment.device import AsyncModbusDevice

logger = get_logger(__name__)


@dataclass
class DeviceGroup:
    """
    設備群組（純順序讀取器）

    依序讀取群組內設備，確保不會同時讀取造成衝突。
    連線/斷線/重連由各 Device 自己管理，DeviceGroup 不介入。

    Attributes:
        devices: 群組內設備列表
        interval: 完整讀取一輪的間隔時間（秒）
        step_interval: 設備間的讀取間隔時間（秒）

    使用範例：
        group = DeviceGroup(
            devices=[device_1, device_2, device_3],
            interval=1.0,
        )
        group.start()
        # ... 運行中 ...
        await group.stop()
    """

    devices: list[AsyncModbusDevice]
    interval: float = 1.0
    step_interval: float = 0.05

    _task: asyncio.Task[None] | None = field(default=None, init=False, repr=False)
    _stop_event: asyncio.Event = field(default_factory=asyncio.Event, init=False, repr=False)

    # ================ 生命週期 ================

    def start(self) -> None:
        """
        啟動順序讀取循環

        建立背景任務，依序讀取群組內所有設備。
        各設備的連線由 Device 自己管理。
        """
        if self._task is not None and not self._task.done():
            return

        self._stop_event.clear()
        self._task = asyncio.create_task(
            self._sequential_loop(),
            name=f"device_group_loop_{id(self)}",
        )
        logger.info(f"設備群組讀取循環已啟動，間隔 {self.interval}s，包含 {len(self.devices)} 個設備")

    async def stop(self) -> None:
        """
        停止順序讀取循環

        取消背景任務並等待其結束。
        """
        self._stop_event.set()

        if self._task is not None and not self._task.done():
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
            self._task = None

        logger.info("設備群組讀取循環已停止")

    # ================ 私有方法 ================

    async def _sequential_loop(self) -> None:
        """
        順序讀取循環

        依序呼叫每個設備的 read_once()，確保不會同時讀取。
        單一設備的讀取錯誤不會影響其他設備。
        """
        while not self._stop_event.is_set():
            start = time.monotonic()

            for device in self.devices:
                if self._stop_event.is_set():
                    break
                if not device.should_attempt_read:
                    logger.debug(f"跳過無回應設備 {device.device_id}，等待重試間隔")
                    continue
                try:
                    await device.read_once()
                except Exception:
                    # read_once 內部已處理錯誤並發射事件
                    # 這裡捕獲避免影響其他設備
                    pass
                finally:
                    await asyncio.sleep(self.step_interval)

            elapsed = time.monotonic() - start
            sleep_time = max(0, self.interval - elapsed)

            if sleep_time > 0:
                try:
                    await asyncio.wait_for(
                        self._stop_event.wait(),
                        timeout=sleep_time,
                    )
                except asyncio.TimeoutError:
                    pass  # 正常超時，繼續下一輪

    # ================ 屬性 ================

    @property
    def is_running(self) -> bool:
        """讀取循環是否運行中"""
        return self._task is not None and not self._task.done()

    @property
    def device_ids(self) -> list[str]:
        """群組內所有設備的 ID"""
        return [d.device_id for d in self.devices]

    def __len__(self) -> int:
        return len(self.devices)

    def __repr__(self) -> str:
        return f"<DeviceGroup devices={self.device_ids} interval={self.interval}s running={self.is_running}>"



# =============== Manager Command - Manager ===============
#
# 寫入指令管理器
#
# 統一管理外部寫入指令：
#   - 維護設備註冊表
#   - 接收指令 → 記錄 DB → 執行 → 更新結果

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from csp_lib.core import get_logger
from csp_lib.equipment.transport import WriteResult, WriteStatus

from .repository import CommandRepository
from .schema import CommandRecord, CommandSource, CommandStatus, WriteCommand

if TYPE_CHECKING:
    from csp_lib.equipment.device import AsyncModbusDevice

logger = get_logger(__name__)


class WriteCommandManager:
    """
    寫入指令管理器

    統一管理外部寫入指令，提供審計日誌與設備路由功能。

    職責：
        1. 維護設備註冊表（device_id → device）
        2. 接收指令 → 記錄 DB → 執行 → 更新結果
        3. 支援多種指令來源（Redis、gRPC、REST 等）

    Example:
        ```python
        from csp_lib.manager.command import WriteCommandManager, MongoCommandRepository

        repo = MongoCommandRepository(db["commands"])
        manager = WriteCommandManager(repo)

        # 註冊設備
        manager.register_device(device1)
        manager.register_device(device2)

        # 執行指令
        command = WriteCommand(
            device_id="device_001",
            point_name="setpoint",
            value=100,
        )
        result = await manager.execute(command)
        ```
    """

    def __init__(self, repository: CommandRepository) -> None:
        """
        初始化寫入指令管理器

        Args:
            repository: 指令記錄儲存庫
        """
        self._repository = repository
        self._devices: dict[str, AsyncModbusDevice] = {}

    # ================ 設備註冊 ================

    def subscribe(self, device: AsyncModbusDevice) -> None:
        """
        註冊可寫入的設備（統一訂閱 API）

        這是所有子管理器的標準訂閱介面。內部委派至 ``register_device()``。

        Args:
            device: Modbus 設備實例
        """
        self.register_device(device)

    def register_device(self, device: AsyncModbusDevice) -> None:
        """
        註冊可寫入的設備

        .. deprecated::
            請改用 ``subscribe()``，此方法保留作為向後相容別名。

        Args:
            device: Modbus 設備實例
        """
        device_id = device.device_id
        if device_id in self._devices:
            logger.warning(f"設備 {device_id} 已註冊，將覆蓋")
        self._devices[device_id] = device
        logger.info(f"寫入指令管理器: 已註冊設備 {device_id}")

    def unregister_device(self, device_id: str) -> None:
        """
        取消註冊設備

        Args:
            device_id: 設備 ID
        """
        if device_id in self._devices:
            del self._devices[device_id]
            logger.info(f"寫入指令管理器: 已取消註冊設備 {device_id}")

    def get_device(self, device_id: str) -> AsyncModbusDevice | None:
        """
        取得已註冊的設備

        Args:
            device_id: 設備 ID

        Returns:
            設備實例，不存在時返回 None
        """
        return self._devices.get(device_id)

    @property
    def registered_device_ids(self) -> list[str]:
        """已註冊的設備 ID 列表"""
        return list(self._devices.keys())

    # ================ 指令執行 ================

    async def execute(self, command: WriteCommand) -> WriteResult:
        """
        執行寫入指令

        流程：
            1. 建立 pending 記錄至 DB
            2. 查找目標設備
            3. 執行寫入
            4. 更新 DB 結果
            5. 回傳結果

        Args:
            command: 寫入指令

        Returns:
            寫入結果
        """
        # 1. 建立記錄
        record = CommandRecord.from_command(command)
        await self._repository.create(record)

        # 2. 查找設備
        device = self._devices.get(command.device_id)
        if device is None:
            await self._repository.update_status(
                command.command_id,
                CommandStatus.DEVICE_NOT_FOUND,
                error_message=f"設備 {command.device_id} 未註冊",
            )
            logger.warning(f"寫入指令失敗: 設備 {command.device_id} 未註冊")
            return WriteResult(
                status=WriteStatus.WRITE_FAILED,
                point_name=command.point_name,
                value=command.value,
                error_message=f"設備 {command.device_id} 未註冊",
            )

        # 3. 更新狀態為執行中
        await self._repository.update_status(command.command_id, CommandStatus.EXECUTING)

        # 4. 執行寫入
        try:
            result = await device.write(
                name=command.point_name,
                value=command.value,
                verify=command.verify,
            )
        except Exception as e:
            await self._repository.update_status(
                command.command_id,
                CommandStatus.FAILED,
                error_message=str(e),
            )
            logger.error(f"寫入指令執行異常: {command.command_id}, {e}")
            return WriteResult(
                status=WriteStatus.WRITE_FAILED,
                point_name=command.point_name,
                value=command.value,
                error_message=str(e),
            )

        # 5. 更新結果
        result_dict = {
            "status": result.status.value,
            "point_name": result.point_name,
            "value": result.value,
            "error_message": result.error_message,
        }

        if result.status == WriteStatus.SUCCESS:
            await self._repository.update_status(
                command.command_id,
                CommandStatus.SUCCESS,
                result=result_dict,
            )
            logger.info(f"寫入指令成功: {command.command_id}")
        else:
            await self._repository.update_status(
                command.command_id,
                CommandStatus.FAILED,
                result=result_dict,
                error_message=result.error_message,
            )
            logger.warning(f"寫入指令失敗: {command.command_id}, {result.error_message}")

        return result

    async def execute_from_dict(
        self,
        data: dict[str, Any],
        source: CommandSource = CommandSource.INTERNAL,
    ) -> WriteResult:
        """
        從字典建立指令並執行

        供 Adapter 使用的便捷方法。

        Args:
            data: 包含 device_id, point_name, value 等欄位的字典
            source: 指令來源

        Returns:
            寫入結果
        """
        command = WriteCommand.from_dict(data, source=source)
        return await self.execute(command)


__all__ = [
    "WriteCommandManager",
]



# =============== Core - Health ===============
#
# 健康檢查介面
#
# 提供統一的健康狀態報告：
#   - HealthStatus: 健康狀態枚舉
#   - HealthReport: 健康報告資料
#   - HealthCheckable: 健康檢查協定

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Protocol, runtime_checkable


class HealthStatus(Enum):
    """健康狀態"""

    HEALTHY = "healthy"
    DEGRADED = "degraded"
    UNHEALTHY = "unhealthy"


@dataclass(frozen=True)
class HealthReport:
    """健康報告"""

    status: HealthStatus
    component: str
    message: str = ""
    details: dict[str, Any] = field(default_factory=dict)
    children: list[HealthReport] = field(default_factory=list)


@runtime_checkable
class HealthCheckable(Protocol):
    """健康檢查協定"""

    def health(self) -> HealthReport: ...


__all__ = [
    "HealthStatus",
    "HealthReport",
    "HealthCheckable",
]



"""Command and write API endpoints."""

from __future__ import annotations

from typing import Any

from fastapi import APIRouter, Depends, HTTPException
from pydantic import BaseModel

from csp_lib.integration import DeviceRegistry, SystemController

from ..dependencies import get_registry, get_system_controller

router = APIRouter(tags=["commands"])


class WriteRequest(BaseModel):
    point_name: str
    value: Any
    verify: bool = False


@router.get("/devices/{device_id}/write-points")
def list_write_points(
    device_id: str,
    registry: DeviceRegistry = Depends(get_registry),
) -> list[str]:
    """List available write point names for a device."""
    device = registry.get_device(device_id)
    if device is None:
        raise HTTPException(status_code=404, detail=f"Device '{device_id}' not found")
    return sorted(device._write_points.keys())


@router.post("/devices/{device_id}/write")
async def write_to_device(
    device_id: str,
    body: WriteRequest,
    registry: DeviceRegistry = Depends(get_registry),
) -> dict[str, Any]:
    """Write a value to a device point."""
    device = registry.get_device(device_id)
    if device is None:
        raise HTTPException(status_code=404, detail=f"Device '{device_id}' not found")

    result = await device.write(body.point_name, body.value, verify=body.verify)
    return {
        "status": result.status.name,
        "point_name": result.point_name,
        "value": result.value,
        "error_message": result.error_message,
    }


@router.post("/executor/trigger")
def trigger_executor(
    sc: SystemController = Depends(get_system_controller),
) -> dict[str, str]:
    """Manually trigger strategy execution."""
    sc.trigger()
    return {"status": "triggered"}



# =============== Manager Schedule - Matcher ===============
#
# 排程時間匹配邏輯
#
# 從 MongoScheduleRepository 提取的純函式，供所有 Repository 實作共用：
#   - matches_time: 檢查當前時間是否在排程時段內（支援跨午夜）
#   - matches_schedule: 檢查排程類型條件是否匹配

from __future__ import annotations

from datetime import date

from .schema import ScheduleRule, ScheduleType


def matches_time(rule: ScheduleRule, now_time_str: str) -> bool:
    """檢查時間是否匹配

    支援跨午夜：若 start_time > end_time（如 22:00-06:00），
    匹配 now >= start_time OR now <= end_time。

    Args:
        rule: 排程規則
        now_time_str: 當前時間（"HH:MM" 格式）

    Returns:
        是否匹配
    """
    start = rule.start_time
    end = rule.end_time

    if start <= end:
        # 正常時段：start <= now <= end
        return start <= now_time_str <= end
    else:
        # 跨午夜：22:00-06:00 → now >= 22:00 OR now <= 06:00
        return now_time_str >= start or now_time_str <= end


def matches_schedule(rule: ScheduleRule, now_weekday: int, now_date: date) -> bool:
    """檢查排程類型條件是否匹配

    Args:
        rule: 排程規則
        now_weekday: 當前星期幾（0=Mon..6=Sun）
        now_date: 當前日期

    Returns:
        是否匹配
    """
    if rule.schedule_type == ScheduleType.DAILY:
        return True
    elif rule.schedule_type == ScheduleType.WEEKLY:
        return now_weekday in rule.days_of_week
    elif rule.schedule_type == ScheduleType.ONCE:
        if rule.start_date and now_date < rule.start_date:
            return False
        if rule.end_date and now_date > rule.end_date:
            return False
        return True
    return False


__all__ = [
    "matches_schedule",
    "matches_time",
]



# =============== Equipment Processing Module ===============
#
# 處理模組匯出

from .aggregator import (
    AggregatorPipeline,
    CoilToBitmaskAggregator,
    ComputedValueAggregator,
)
from .can_encoder import CANFieldEncoder, CANFrameBuffer, CANSignalDefinition, FrameBufferConfig
from .can_parser import CANField, CANFrameParser
from .decoder import ModbusDecoder, ModbusEncoder

__all__ = [
    # Decoder/Encoder
    "ModbusDecoder",
    "ModbusEncoder",
    # Aggregator
    "CoilToBitmaskAggregator",
    "ComputedValueAggregator",
    "AggregatorPipeline",
    # CAN Parser
    "CANField",
    "CANFrameParser",
    # CAN Encoder
    "CANSignalDefinition",
    "FrameBufferConfig",
    "CANFieldEncoder",
    "CANFrameBuffer",
]



# =============== Controller Module ===============
#
# 控制器模組頂層匯出
#
# 提供便捷的 import 路徑：
#   from csp_lib.controller import Strategy, Command, StrategyExecutor

from .calibration import FFCalibrationConfig, FFCalibrationStrategy
from .compensator import PowerCompensator, PowerCompensatorConfig
from .core import (
    Command,
    CommandProcessor,
    ConfigMixin,
    ExecutionConfig,
    ExecutionMode,
    Strategy,
    StrategyContext,
    SystemBase,
)
from .discovery import ENTRY_POINT_GROUP, StrategyDescriptor, discover_strategies
from .executor import ComputeOffloader, StrategyExecutor
from .protocol import GridControllerBase, GridControllerProtocol
from .services import PVDataService
from .strategies import (
    BypassStrategy,
    DroopConfig,
    DroopStrategy,
    FPConfig,
    FPStrategy,
    IslandModeConfig,
    IslandModeStrategy,
    LoadCircuitProtocol,
    LoadSheddingConfig,
    LoadSheddingStrategy,
    PQModeConfig,
    PQModeStrategy,
    PVSmoothConfig,
    PVSmoothStrategy,
    QVConfig,
    QVStrategy,
    RampStopStrategy,
    RelayProtocol,
    RemainingTimeCondition,
    ScheduleStrategy,
    ShedCondition,
    ShedStage,
    StopStrategy,
    ThresholdCondition,
)
from .system import (
    AlarmStopOverride,
    CapacityConfig,
    CascadingStrategy,
    ContextKeyOverride,
    DynamicSOCProtection,
    EventDrivenOverride,
    GridLimitProtection,
    ModeDefinition,
    ModeManager,
    ModePriority,
    ProtectionGuard,
    ProtectionResult,
    ProtectionRule,
    RampStopProtection,
    ReversePowerProtection,
    ScheduleModeController,
    SOCProtection,
    SOCProtectionConfig,
    SwitchSource,
    SystemAlarmProtection,
)

__all__ = [
    # Calibration
    "FFCalibrationConfig",
    "FFCalibrationStrategy",
    # Compensator
    "PowerCompensator",
    "PowerCompensatorConfig",
    # Discovery
    "ENTRY_POINT_GROUP",
    "StrategyDescriptor",
    "discover_strategies",
    # Protocol
    "GridControllerBase",
    "GridControllerProtocol",
    # Core
    "Command",
    "CommandProcessor",
    "SystemBase",
    "ConfigMixin",
    "StrategyContext",
    "ExecutionMode",
    "ExecutionConfig",
    "Strategy",
    # Executor
    "ComputeOffloader",
    "StrategyExecutor",
    # Services
    "PVDataService",
    # Strategies
    "BypassStrategy",
    "DroopConfig",
    "DroopStrategy",
    "FPConfig",
    "FPStrategy",
    "IslandModeConfig",
    "IslandModeStrategy",
    "LoadCircuitProtocol",
    "LoadSheddingConfig",
    "LoadSheddingStrategy",
    "PQModeConfig",
    "PQModeStrategy",
    "PVSmoothConfig",
    "PVSmoothStrategy",
    "QVConfig",
    "QVStrategy",
    "RelayProtocol",
    "RampStopStrategy",
    "RemainingTimeCondition",
    "ScheduleStrategy",
    "ShedCondition",
    "ShedStage",
    "StopStrategy",
    "ThresholdCondition",
    # System
    "AlarmStopOverride",
    "CapacityConfig",
    "CascadingStrategy",
    "ContextKeyOverride",
    "DynamicSOCProtection",
    "EventDrivenOverride",
    "GridLimitProtection",
    "RampStopProtection",
    "ScheduleModeController",
    "SwitchSource",
    "ModePriority",
    "ModeDefinition",
    "ModeManager",
    "ProtectionRule",
    "SOCProtection",
    "SOCProtectionConfig",
    "ReversePowerProtection",
    "SystemAlarmProtection",
    "ProtectionResult",
    "ProtectionGuard",
]



# =============== Manager Command - Redis Adapter ===============
#
# Redis Pub/Sub 指令適配器
#
# 監聽 Redis channel 接收指令並轉發至 WriteCommandManager
# 支援兩種指令類型：
#   - WriteCommand: 點位寫入 {"device_id", "point_name", "value"}
#   - ActionCommand: 動作執行 {"device_id", "action", "params"}

from __future__ import annotations

import asyncio
import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from csp_lib.core import AsyncLifecycleMixin, get_logger
from csp_lib.equipment.transport import WriteStatus

from ..schema import ActionCommand, CommandSource
from .config import CommandAdapterConfig

if TYPE_CHECKING:
    from redis.asyncio import Redis

    from csp_lib.integration.orchestrator import SystemCommandOrchestrator

    from ..manager import WriteCommandManager

logger = get_logger(__name__)


# ========== Result Schema ==========


@dataclass(frozen=True)
class CommandResult:
    """
    指令執行結果

    用於發布到 result_channel 的標準化結果格式。

    Attributes:
        command_id: 指令 ID
        device_id: 設備 ID
        status: 執行狀態
        action: 動作名稱（ActionCommand 時）
        point_name: 點位名稱（WriteCommand 時）
        value: 寫入值或動作參數
        error_message: 錯誤訊息
    """

    command_id: str
    device_id: str
    status: str
    action: str | None = None
    point_name: str | None = None
    value: Any = None
    error_message: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """轉為字典（用於 JSON 序列化）"""
        return {
            "command_id": self.command_id,
            "device_id": self.device_id,
            "status": self.status,
            "action": self.action,
            "point_name": self.point_name,
            "value": self.value,
            "error_message": self.error_message,
        }


# ========== Redis Adapter ==========


class RedisCommandAdapter(AsyncLifecycleMixin):
    """
    Redis Pub/Sub 指令適配器

    監聯指定 channel，解析 JSON 指令並轉發至 WriteCommandManager。
    執行完成後將結果發布到結果 channel。

    支援兩種指令類型：
        - 點位寫入 (WriteCommand): {"device_id", "point_name", "value", "verify"}
        - 動作執行 (ActionCommand): {"device_id", "action", "params"}

    Attributes:
        _redis: Redis 客戶端
        _manager: 寫入指令管理器
        _command_channel: 接收指令的 channel
        _result_channel: 發布結果的 channel

    Example:
        ```python
        adapter = RedisCommandAdapter(
            redis_client=redis_client._client,
            manager=command_manager,
        )
        await adapter.start()

        # 點位寫入
        # PUBLISH channel:commands:write '{"device_id":"d1","point_name":"sp","value":100}'

        # 動作執行
        # PUBLISH channel:commands:write '{"device_id":"Generator","action":"start"}'
        # PUBLISH channel:commands:write '{"device_id":"Generator","action":"set_power","params":{"p":80}}'
        ```
    """

    def __init__(
        self,
        redis_client: Redis,
        manager: WriteCommandManager,
        config: CommandAdapterConfig | None = None,
        command_channel: str | None = None,
        result_channel: str | None = None,
        orchestrator: SystemCommandOrchestrator | None = None,
    ) -> None:
        """
        初始化 Redis 指令適配器

        Args:
            redis_client: redis.asyncio.Redis 客戶端實例
            manager: 寫入指令管理器
            config: 適配器配置（優先使用）
            command_channel: 接收指令的 channel，config 為 None 時使用
            result_channel: 發布結果的 channel，config 為 None 時使用
            orchestrator: 系統指令編排器（可選，提供時支援系統級指令）
        """
        self._redis = redis_client
        self._manager = manager
        if config is None:
            config = CommandAdapterConfig(
                command_channel=command_channel or "channel:commands:write",
                result_channel=result_channel or "channel:commands:result",
            )
        self._config = config
        self._command_channel = self._config.command_channel
        self._result_channel = self._config.result_channel
        self._orchestrator = orchestrator
        self._running = False
        self._task: asyncio.Task | None = None

    @property
    def is_running(self) -> bool:
        """是否正在運行"""
        return self._running

    async def _on_start(self) -> None:
        """
        啟動監聽

        開始訂閱 command_channel 並處理指令。
        """
        if self._running:
            return

        self._running = True
        self._task = asyncio.create_task(self._listen_loop())
        logger.info(f"Redis 指令適配器已啟動: {self._command_channel}")

    async def _on_stop(self) -> None:
        """
        停止監聽

        取消訂閱並停止處理。
        """
        if not self._running:
            return

        self._running = False
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
            self._task = None
        logger.info("Redis 指令適配器已停止")

    async def _listen_loop(self) -> None:
        """監聽循環"""
        pubsub = self._redis.pubsub()
        await pubsub.subscribe(self._command_channel)

        try:
            while self._running:
                message = await pubsub.get_message(ignore_subscribe_messages=True, timeout=1.0)
                if message and message["type"] == "message":
                    await self._handle_message(message["data"])
        except asyncio.CancelledError:
            pass
        finally:
            await pubsub.unsubscribe(self._command_channel)
            await pubsub.aclose()

    async def _handle_message(self, data: str | bytes) -> None:
        """處理訊息"""
        try:
            if isinstance(data, bytes):
                data = data.decode("utf-8")

            command_data = json.loads(data)
            logger.debug(f"收到指令: {command_data}")

            # 判斷指令類型：系統指令 → 動作指令 → 寫入指令
            if "system_command" in command_data:
                result = await self._execute_system_command(command_data)
            elif ActionCommand.is_action_command(command_data):
                result = await self._execute_action(command_data)
            else:
                result = await self._execute_write(command_data)

            # 發布結果
            await self._publish_result(result)

        except json.JSONDecodeError as e:
            logger.error(f"指令 JSON 解析失敗: {e}")
        except KeyError as e:
            logger.error(f"指令缺少必要欄位: {e}")
        except Exception as e:
            logger.error(f"指令處理失敗: {e}")

    async def _execute_action(self, data: dict[str, Any]) -> CommandResult:
        """
        執行動作指令

        Args:
            data: 包含 device_id, action, value(可選) 的字典

        Returns:
            CommandResult
        """
        try:
            command = ActionCommand.from_dict(data, source=CommandSource.REDIS_PUBSUB)
        except KeyError as e:
            return CommandResult(
                command_id=data.get("command_id", ""),
                device_id=data.get("device_id", ""),
                status=WriteStatus.VALIDATION_FAILED.value,
                action=data.get("action"),
                error_message=f"Missing required field: {e}",
            )

        device = self._manager.get_device(command.device_id)
        if device is None:
            return CommandResult(
                command_id=command.command_id,
                device_id=command.device_id,
                status=WriteStatus.WRITE_FAILED.value,
                action=command.action,
                error_message=f"Device '{command.device_id}' not registered",
            )

        logger.info(f"執行 action: {command.device_id}.{command.action}, value={command.value}")
        result = await device.execute_action(command.action, **command.params)
        if result.status != WriteStatus.SUCCESS:
            logger.error(
                f"Action 執行失敗: device_id={command.device_id}, action={command.action}, value={command.value}, error={result.error_message}"
            )
            return CommandResult(
                command_id=command.command_id,
                device_id=command.device_id,
                status=result.status.value,
                action=command.action,
                value=command.value,
                error_message=result.error_message,
            )
        else:
            logger.info(
                f"Action 執行成功: device_id={command.device_id}, action={command.action}, value={command.value}"
            )

        return CommandResult(
            command_id=command.command_id,
            device_id=command.device_id,
            status=result.status.value,
            action=command.action,
            value=command.value,
            error_message=result.error_message,
        )

    async def _execute_system_command(self, data: dict[str, Any]) -> CommandResult:
        """
        執行系統級指令

        Args:
            data: 包含 system_command 的字典

        Returns:
            CommandResult
        """
        command_name = data["system_command"]

        if self._orchestrator is None:
            return CommandResult(
                command_id=data.get("command_id", ""),
                device_id="system",
                status="failed",
                action=command_name,
                error_message="System command orchestrator is not configured",
            )

        try:
            result = await self._orchestrator.execute(command_name)
            return CommandResult(
                command_id=data.get("command_id", ""),
                device_id="system",
                status=result.status,
                action=command_name,
                error_message=result.error_message,
            )
        except KeyError as e:
            return CommandResult(
                command_id=data.get("command_id", ""),
                device_id="system",
                status="failed",
                action=command_name,
                error_message=str(e),
            )

    async def _execute_write(self, data: dict[str, Any]) -> CommandResult:
        """
        執行寫入指令

        Args:
            data: 包含 device_id, point_name, value 的字典

        Returns:
            CommandResult
        """
        result = await self._manager.execute_from_dict(
            data,
            source=CommandSource.REDIS_PUBSUB,
        )

        return CommandResult(
            command_id=data.get("command_id", ""),
            device_id=data.get("device_id", ""),
            status=result.status.value,
            point_name=result.point_name,
            value=result.value,
            error_message=result.error_message,
        )

    async def _publish_result(self, result: CommandResult) -> None:
        """
        發布執行結果

        Args:
            result: 指令執行結果
        """
        message = json.dumps(result.to_dict())
        await self._redis.publish(self._result_channel, message)


__all__ = [
    "RedisCommandAdapter",
    "CommandResult",
]



"""
RuntimeParameters — Thread-safe 即時參數容器

提供跨執行緒 / 跨 asyncio task 安全的參數讀寫，用於：
- ProtectionRule 讀取動態 SOC 上下限、功率限制
- ModbusGatewayServer WriteHook 寫入 EMS 指令
- ContextBuilder 注入 context.extra
- 任何需要從外部系統即時更新的參數

使用 threading.Lock 確保 Modbus thread (modbus_tk hook) 與
asyncio event loop 之間的安全存取。

Usage::

    params = RuntimeParameters(
        soc_max=95.0,
        soc_min=5.0,
        grid_limit_pct=100,
    )

    # 讀取
    soc_max = params.get("soc_max")

    # 寫入（觸發 observers）
    params.set("soc_max", 90.0)

    # 批次更新
    params.update({"soc_max": 90.0, "soc_min": 10.0})

    # 原子性快照
    snap = params.snapshot()

    # 變更通知
    params.on_change(lambda key, old, new: print(f"{key}: {old} → {new}"))
"""

from __future__ import annotations

import threading
from typing import Any, Callable

from loguru import logger as _root_logger

logger = _root_logger.bind(module="csp_lib.core.runtime_params")

# observer 簽名: (key, old_value, new_value) -> None
ChangeCallback = Callable[[str, Any, Any], None]


class RuntimeParameters:
    """
    Thread-safe 的即時參數容器

    所有讀寫操作透過 threading.Lock 保護，確保：
    - asyncio task 之間的一致性
    - Modbus hook thread 與 asyncio event loop 之間的安全存取

    Attributes:
        _values: 參數鍵值對
        _lock: 保護 _values 的鎖
        _observers: 變更通知回呼列表
    """

    __slots__ = ("_values", "_lock", "_observers")

    def __init__(self, **initial_values: Any) -> None:
        self._lock = threading.Lock()
        self._values: dict[str, Any] = dict(initial_values)
        self._observers: list[ChangeCallback] = []

    # ─────────────────────── 讀取 ───────────────────────

    def get(self, key: str, default: Any = None) -> Any:
        """取得參數值，不存在時回傳 default"""
        with self._lock:
            return self._values.get(key, default)

    def snapshot(self) -> dict[str, Any]:
        """回傳所有參數的原子性淺拷貝"""
        with self._lock:
            return dict(self._values)

    def keys(self) -> list[str]:
        """回傳所有參數 key"""
        with self._lock:
            return list(self._values.keys())

    def __contains__(self, key: str) -> bool:
        with self._lock:
            return key in self._values

    def __len__(self) -> int:
        with self._lock:
            return len(self._values)

    # ─────────────────────── 寫入 ───────────────────────

    def set(self, key: str, value: Any) -> None:
        """設定單一參數值，值變更時觸發 observers"""
        with self._lock:
            old = self._values.get(key)
            self._values[key] = value
        if old != value:
            self._notify(key, old, value)

    def update(self, mapping: dict[str, Any]) -> None:
        """批次更新多個參數，值變更時觸發 observers"""
        changes: list[tuple[str, Any, Any]] = []
        with self._lock:
            for key, value in mapping.items():
                old = self._values.get(key)
                self._values[key] = value
                if old != value:
                    changes.append((key, old, value))
        for key, old, new in changes:
            self._notify(key, old, new)

    def setdefault(self, key: str, default: Any) -> Any:
        """若 key 不存在則設定為 default 並回傳，存在則直接回傳"""
        with self._lock:
            if key not in self._values:
                self._values[key] = default
                return default
            return self._values[key]

    def delete(self, key: str) -> None:
        """刪除參數，不存在時靜默忽略"""
        with self._lock:
            old = self._values.pop(key, None)
        if old is not None:
            self._notify(key, old, None)

    # ─────────────────────── 觀察者 ───────────────────────

    def on_change(self, callback: ChangeCallback) -> None:
        """
        註冊變更通知回呼

        回呼簽名: callback(key: str, old_value: Any, new_value: Any) -> None
        回呼在 set/update/delete 時同步呼叫（在鎖外），
        若需要 async 操作，請在回呼內使用 loop.call_soon_threadsafe()。
        """
        self._observers.append(callback)

    def remove_observer(self, callback: ChangeCallback) -> None:
        """移除已註冊的變更通知回呼"""
        try:
            self._observers.remove(callback)
        except ValueError:
            pass

    def _notify(self, key: str, old: Any, new: Any) -> None:
        """通知所有觀察者，單一觀察者例外不影響其他"""
        for cb in self._observers:
            try:
                cb(key, old, new)
            except Exception:
                logger.opt(exception=True).warning(f"RuntimeParameters observer 執行失敗: key={key}")

    # ─────────────────────── 表示 ───────────────────────

    def __repr__(self) -> str:
        with self._lock:
            keys = list(self._values.keys())
        return f"RuntimeParameters({', '.join(keys)})"


__all__ = ["RuntimeParameters", "ChangeCallback"]



# =============== Core - Lifecycle ===============
#
# Async 生命週期 Mixin
#
# 提供標準的 start/stop 與 async context manager 框架：
#   - AsyncLifecycleMixin: 子類覆寫 _on_start/_on_stop 即可

from __future__ import annotations

from typing import Any, Self


class AsyncLifecycleMixin:
    """
    Async 生命週期 Mixin

    提供標準的 start/stop 與 async context manager 實作框架。
    子類別只需覆寫 ``_on_start()`` 與 ``_on_stop()`` 即可。

    使用範例::

        class MyService(AsyncLifecycleMixin):
            async def _on_start(self) -> None:
                # 啟動邏輯
                ...

            async def _on_stop(self) -> None:
                # 停止邏輯
                ...

        async with MyService() as svc:
            ...
    """

    async def start(self) -> None:
        """啟動服務"""
        await self._on_start()

    async def stop(self) -> None:
        """停止服務"""
        await self._on_stop()

    async def _on_start(self) -> None:
        """子類別覆寫此方法以實作啟動邏輯"""

    async def _on_stop(self) -> None:
        """子類別覆寫此方法以實作停止邏輯"""

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.stop()


__all__ = [
    "AsyncLifecycleMixin",
]



# =============== Manager Schedule - Schema ===============
#
# 排程規則資料結構定義
#
# 提供排程系統的核心資料模型：
#   - ScheduleType: 排程類型枚舉（單次/每日/每週）
#   - StrategyType: 策略類型枚舉
#   - ScheduleRule: 排程規則資料類別（MongoDB Document）

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date
from enum import Enum
from typing import Any


class ScheduleType(str, Enum):
    """
    排程類型枚舉

    Values:
        ONCE: 單次排程（指定日期範圍）
        DAILY: 每日排程（每天重複）
        WEEKLY: 每週排程（指定星期幾）
    """

    ONCE = "once"
    DAILY = "daily"
    WEEKLY = "weekly"


class StrategyType(str, Enum):
    """
    策略類型枚舉

    Values:
        PQ: 固定功率模式
        PV_SMOOTH: PV 平滑模式
        QV: 電壓無功控制
        FP: 頻率功率控制
        ISLAND: 離網模式
        BYPASS: 旁路模式
        STOP: 停止模式
    """

    PQ = "pq"
    PV_SMOOTH = "pv_smooth"
    QV = "qv"
    FP = "fp"
    ISLAND = "island"
    BYPASS = "bypass"
    STOP = "stop"


@dataclass
class ScheduleRule:
    """
    排程規則資料類別

    對應 MongoDB Document，用於儲存單筆排程規則的完整資訊。

    Attributes:
        name: 規則名稱
        site_id: 站點識別碼
        schedule_type: 排程類型（ONCE/DAILY/WEEKLY）
        strategy_type: 策略類型
        strategy_config: 策略配置字典
        start_time: 開始時間（"HH:MM" 格式）
        end_time: 結束時間（"HH:MM" 格式）
        priority: 優先順序（數字越大優先級越高）
        enabled: 是否啟用
        days_of_week: 每週排程的星期幾列表（0=Mon..6=Sun）
        start_date: 單次排程的開始日期
        end_date: 單次排程的結束日期
    """

    name: str
    site_id: str
    schedule_type: ScheduleType
    strategy_type: StrategyType
    strategy_config: dict[str, Any] = field(default_factory=dict)
    start_time: str = "00:00"
    end_time: str = "23:59"
    priority: int = 0
    enabled: bool = True
    days_of_week: list[int] = field(default_factory=list)
    start_date: date | None = None
    end_date: date | None = None

    def to_document(self) -> dict[str, Any]:
        """
        轉換為 MongoDB document

        Enum 欄位轉換為 value 字串，date 欄位轉換為 ISO 字串。

        Returns:
            dict[str, Any]: MongoDB document
        """
        doc: dict[str, Any] = {
            "name": self.name,
            "site_id": self.site_id,
            "type": self.schedule_type.value,
            "strategy_type": self.strategy_type.value,
            "strategy_config": self.strategy_config,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "priority": self.priority,
            "enabled": self.enabled,
            "days_of_week": self.days_of_week,
            "start_date": self.start_date.isoformat() if self.start_date else None,
            "end_date": self.end_date.isoformat() if self.end_date else None,
        }
        return doc

    @classmethod
    def from_document(cls, doc: dict[str, Any]) -> ScheduleRule:
        """
        從 MongoDB document 建立 ScheduleRule

        自動處理 _id 移除、Enum 轉換、日期解析。

        Args:
            doc: MongoDB document

        Returns:
            ScheduleRule: 排程規則物件
        """
        doc = doc.copy()
        doc.pop("_id", None)

        # DB field "type" -> Python field "schedule_type"
        schedule_type_value = doc.pop("type")

        # 解析 date 欄位
        start_date = None
        if doc.get("start_date"):
            start_date = date.fromisoformat(doc["start_date"])

        end_date = None
        if doc.get("end_date"):
            end_date = date.fromisoformat(doc["end_date"])

        return cls(
            name=doc["name"],
            site_id=doc["site_id"],
            schedule_type=ScheduleType(schedule_type_value),
            strategy_type=StrategyType(doc["strategy_type"]),
            strategy_config=doc.get("strategy_config", {}),
            start_time=doc.get("start_time", "00:00"),
            end_time=doc.get("end_time", "23:59"),
            priority=doc.get("priority", 0),
            enabled=doc.get("enabled", True),
            days_of_week=doc.get("days_of_week", []),
            start_date=start_date,
            end_date=end_date,
        )


__all__ = [
    "ScheduleRule",
    "ScheduleType",
    "StrategyType",
]



# =============== Equipment Transport - Periodic Sender ===============
#
# 定期發送排程器
#
# 為每個 CAN ID 建立獨立的 asyncio.Task，按週期發送 buffer 最新內容。

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Awaitable, Callable

from csp_lib.core import get_logger

if TYPE_CHECKING:
    from csp_lib.equipment.processing.can_encoder import CANFrameBuffer

logger = get_logger(__name__)


@dataclass(frozen=True)
class PeriodicFrameConfig:
    """
    定期發送配置

    Attributes:
        can_id: CAN 訊框 ID
        interval: 發送間隔（秒）
        enabled: 是否啟用
    """

    can_id: int
    interval: float = 0.1
    enabled: bool = True


class PeriodicSendScheduler:
    """
    定期發送排程器

    為每個 CAN ID 建立獨立的 asyncio.Task，按週期從 frame_buffer 取得最新內容並發送。
    支持暫停/恢復/立即發送。

    使用範例::

        scheduler = PeriodicSendScheduler(
            frame_buffer=buffer,
            send_callback=client.send,
            configs=[PeriodicFrameConfig(can_id=0x200, interval=0.1)],
        )
        await scheduler.start()
        # ... 使用中 ...
        await scheduler.stop()
    """

    def __init__(
        self,
        frame_buffer: CANFrameBuffer,
        send_callback: Callable[[int, bytes], Awaitable[None]],
        configs: list[PeriodicFrameConfig],
    ) -> None:
        self._frame_buffer = frame_buffer
        self._send_callback = send_callback
        self._configs = {cfg.can_id: cfg for cfg in configs}
        self._tasks: dict[int, asyncio.Task[None]] = {}
        self._paused: set[int] = set()
        self._running = False

    async def start(self) -> None:
        """啟動所有定期發送任務"""
        if self._running:
            return
        self._running = True
        for can_id, cfg in self._configs.items():
            if cfg.enabled:
                self._tasks[can_id] = asyncio.create_task(
                    self._send_loop(can_id, cfg.interval),
                    name=f"periodic_sender_0x{can_id:03X}",
                )

    async def stop(self) -> None:
        """停止所有定期發送任務"""
        if not self._running:
            return
        self._running = False
        for task in self._tasks.values():
            task.cancel()
        for task in self._tasks.values():
            try:
                await task
            except asyncio.CancelledError:
                pass
        self._tasks.clear()

    def pause(self, can_id: int) -> None:
        """暫停指定 CAN ID 的定期發送"""
        self._paused.add(can_id)

    def resume(self, can_id: int) -> None:
        """恢復指定 CAN ID 的定期發送"""
        self._paused.discard(can_id)

    async def send_now(self, can_id: int) -> None:
        """立即發送指定 CAN ID 的訊框"""
        data = self._frame_buffer.get_frame(can_id)
        await self._send_callback(can_id, data)

    async def _send_loop(self, can_id: int, interval: float) -> None:
        """單一 CAN ID 的發送循環"""
        while self._running:
            try:
                if can_id not in self._paused:
                    data = self._frame_buffer.get_frame(can_id)
                    await self._send_callback(can_id, data)
                await asyncio.sleep(interval)
            except asyncio.CancelledError:
                break
            except Exception:
                logger.opt(exception=True).warning("定期發送失敗: can_id=0x{:03X}", can_id)
                await asyncio.sleep(interval)


__all__ = [
    "PeriodicFrameConfig",
    "PeriodicSendScheduler",
]



# =============== Dynamic Protection Rules ===============
#
# 從 RuntimeParameters 讀取動態參數的保護規則
#
# 解決 SOCProtectionConfig 等 frozen dataclass 無法即時更新的問題。
# 適用於從外部系統（EMS/Modbus/Redis）即時推送的參數。
#
# 規則：
#   - DynamicSOCProtection: 動態 SOC 上下限保護
#   - GridLimitProtection: 外部功率限制（電力公司/排程）
#   - RampStopProtection: 故障/告警時斜坡降功率至 0

from __future__ import annotations

from typing import TYPE_CHECKING

from csp_lib.controller.core import Command, StrategyContext
from csp_lib.core import get_logger

from .protection import ProtectionRule, SOCProtectionConfig

if TYPE_CHECKING:
    from csp_lib.core.runtime_params import RuntimeParameters

logger = get_logger(__name__)


# =============== Dynamic SOC Protection ===============


class DynamicSOCProtection(ProtectionRule):
    """
    動態 SOC 保護

    支援兩種參數來源：

    1. **RuntimeParameters**（動態）：每次 evaluate() 從 RuntimeParameters 讀取
       soc_max / soc_min，支援即時更新（如來自 EMS/Modbus 的寫入）。
    2. **SOCProtectionConfig**（靜態）：從 frozen dataclass 讀取固定的
       soc_high / soc_low / warning_band，取代已棄用的 SOCProtection。

    P > 0 = 放電；P < 0 = 充電

    RuntimeParameters keys (僅 RuntimeParameters 模式):
        soc_max: SOC 上限（%），預設 95.0
        soc_min: SOC 下限（%），預設 5.0

    Args:
        params: RuntimeParameters 或 SOCProtectionConfig 實例
        soc_max_key: soc_max 在 RuntimeParameters 中的 key（SOCProtectionConfig 模式忽略）
        soc_min_key: soc_min 在 RuntimeParameters 中的 key（SOCProtectionConfig 模式忽略）
        warning_band: 警戒區寬度（%），0 = 不啟用漸進限制（SOCProtectionConfig 模式使用 config 值）
    """

    def __init__(
        self,
        params: RuntimeParameters | SOCProtectionConfig,
        soc_max_key: str = "soc_max",
        soc_min_key: str = "soc_min",
        warning_band: float = 0.0,
    ) -> None:
        self._params = params
        self._soc_max_key = soc_max_key
        self._soc_min_key = soc_min_key
        self._warning_band = warning_band
        self._is_triggered = False

    @property
    def name(self) -> str:
        return "dynamic_soc_protection"

    @property
    def is_triggered(self) -> bool:
        return self._is_triggered

    def _resolve_limits(self) -> tuple[float, float, float]:
        """Resolve soc_max, soc_min, warning_band from the parameter source."""
        if isinstance(self._params, SOCProtectionConfig):
            return self._params.soc_high, self._params.soc_low, self._params.warning_band
        # RuntimeParameters path
        soc_max = float(self._params.get(self._soc_max_key, 95.0))
        soc_min = float(self._params.get(self._soc_min_key, 5.0))
        return soc_max, soc_min, self._warning_band

    def evaluate(self, command: Command, context: StrategyContext) -> Command:
        soc = context.soc
        if soc is None:
            self._is_triggered = False
            return command

        soc_max, soc_min, wb = self._resolve_limits()
        p = command.p_target

        # SOC 過高：禁止充電
        if soc >= soc_max:
            if p < 0:
                self._is_triggered = True
                logger.warning(f"DynamicSOC: SOC={soc:.1f}% >= {soc_max}%, block charging, P: {p:.1f} → 0")
                return command.with_p(0.0)
            self._is_triggered = False
            return command

        # SOC 過低：禁止放電
        if soc <= soc_min:
            if p > 0:
                self._is_triggered = True
                logger.warning(f"DynamicSOC: SOC={soc:.1f}% <= {soc_min}%, block discharging, P: {p:.1f} → 0")
                return command.with_p(0.0)
            self._is_triggered = False
            return command

        # 高側警戒區：漸進限制充電
        if wb > 0 and soc >= soc_max - wb and p < 0:
            ratio = (soc_max - soc) / wb
            limited = p * ratio
            self._is_triggered = True
            logger.debug(f"DynamicSOC: high warning SOC={soc:.1f}%, ratio={ratio:.2f}, P: {p:.1f} → {limited:.1f}")
            return command.with_p(limited)

        # 低側警戒區：漸進限制放電
        if wb > 0 and soc <= soc_min + wb and p > 0:
            ratio = (soc - soc_min) / wb
            limited = p * ratio
            self._is_triggered = True
            logger.debug(f"DynamicSOC: low warning SOC={soc:.1f}%, ratio={ratio:.2f}, P: {p:.1f} → {limited:.1f}")
            return command.with_p(limited)

        self._is_triggered = False
        return command


# =============== Grid Limit Protection ===============


class GridLimitProtection(ProtectionRule):
    """
    外部功率限制保護

    讀取 RuntimeParameters 中的功率限制百分比，計算上限：
        max_p = total_rated_kw × limit_pct / 100

    正負 P 均受限制，clamp 至 [-max_p, +max_p]。

    RuntimeParameters keys:
        grid_limit_pct: 功率限制百分比（0~100），預設 100（無限制）

    Args:
        params: RuntimeParameters 實例
        total_rated_kw: 系統額定功率 (kW)
        limit_key: limit_pct 在 params 中的 key
    """

    def __init__(
        self,
        params: RuntimeParameters,
        total_rated_kw: float,
        limit_key: str = "grid_limit_pct",
    ) -> None:
        self._params = params
        self._rated = total_rated_kw
        self._limit_key = limit_key
        self._is_triggered = False

    @property
    def name(self) -> str:
        return "grid_limit_protection"

    @property
    def is_triggered(self) -> bool:
        return self._is_triggered

    def evaluate(self, command: Command, context: StrategyContext) -> Command:
        pct = float(self._params.get(self._limit_key, 100))
        max_p = self._rated * pct / 100.0
        p = command.p_target

        if p > max_p:
            self._is_triggered = True
            logger.warning(f"GridLimit: discharge over limit={pct:.0f}%, P: {p:.1f} → {max_p:.1f}")
            return command.with_p(max_p)

        if p < -max_p:
            self._is_triggered = True
            logger.warning(f"GridLimit: charge over limit={pct:.0f}%, P: {p:.1f} → {-max_p:.1f}")
            return command.with_p(-max_p)

        self._is_triggered = False
        return command


# =============== Ramp Stop Protection (Deprecated) ===============
# 建議改用 csp_lib.controller.strategies.RampStopStrategy + EventDrivenOverride
# RampStop 本質上是「接管控制」而非「修改數值」，更適合作為 Strategy。


class RampStopProtection(ProtectionRule):
    """
    斜坡停機保護

    當 RuntimeParameters 中的觸發旗標為 True 時，
    從上次保護後的實際功率出發，按斜率逐步降至 0。

    ramp_step = ramp_rate(%) / 100 × total_rated_kw × interval_seconds

    起點為 context.last_command.p_target（上次保護後實際送出的 P）。

    RuntimeParameters keys:
        trigger_key: 觸發旗標 (bool/int)，預設 "battery_status"
        ramp_rate_key: 斜率百分比 (%/s)，預設 "ramp_rate"

    Args:
        params: RuntimeParameters 實例
        total_rated_kw: 系統額定功率 (kW)
        interval_seconds: 控制週期（秒），用於計算每步降幅
        trigger_key: 觸發旗標的 key
        trigger_value: 觸發值（預設 1，與 battery_status=1 對應）
        ramp_rate_key: 斜率的 key
        default_ramp_rate: 無 key 時的預設斜率（%/s）
    """

    def __init__(
        self,
        params: RuntimeParameters,
        total_rated_kw: float,
        interval_seconds: float = 0.3,
        trigger_key: str = "battery_status",
        trigger_value: int = 1,
        ramp_rate_key: str = "ramp_rate",
        default_ramp_rate: float = 5.0,
    ) -> None:
        self._params = params
        self._rated = total_rated_kw
        self._interval = interval_seconds
        self._trigger_key = trigger_key
        self._trigger_value = trigger_value
        self._ramp_rate_key = ramp_rate_key
        self._default_ramp_rate = default_ramp_rate
        self._is_triggered = False

    @property
    def name(self) -> str:
        return "ramp_stop_protection"

    @property
    def is_triggered(self) -> bool:
        return self._is_triggered

    def evaluate(self, command: Command, context: StrategyContext) -> Command:
        trigger = self._params.get(self._trigger_key, 0)
        if trigger != self._trigger_value:
            self._is_triggered = False
            return command

        self._is_triggered = True

        # 以上次保護後的 P 為起點斜坡降至 0
        current_p = context.last_command.p_target
        ramp_rate = float(self._params.get(self._ramp_rate_key, self._default_ramp_rate))
        ramp_step = ramp_rate / 100.0 * self._rated * self._interval

        if abs(current_p) <= ramp_step:
            target_p = 0.0
        elif current_p > 0:
            target_p = current_p - ramp_step
        else:
            target_p = current_p + ramp_step

        logger.debug(
            f"RampStop: triggered, current_p={current_p:.1f}kW, ramp_step={ramp_step:.1f}kW, target={target_p:.1f}kW"
        )
        return command.with_p(target_p)


__all__ = [
    "DynamicSOCProtection",
    "GridLimitProtection",
    "RampStopProtection",
]



# =============== Integration Distributed - Config ===============
#
# 分散式控制配置
#
# 定義遠端站台與分散式控制器的配置：
#   - RemoteSiteConfig: 單一遠端站台設定
#   - DistributedConfig: 分散式控制器總配置

from __future__ import annotations

from dataclasses import dataclass, field


@dataclass(frozen=True)
class RemoteSiteConfig:
    """
    遠端站台配置

    定義一個遠端站台（例如 Computer_1 上的 BMS）的相關設定。

    Attributes:
        site_id: 站台唯一識別碼（例如 "site_bms"）
        device_ids: 此站台上的設備 ID 列表
        command_channel: 指令接收 channel（預設 "channel:commands:{site_id}:write"）
        result_channel: 結果回傳 channel（預設 "channel:commands:{site_id}:result"）
    """

    site_id: str
    device_ids: list[str] = field(default_factory=list)
    command_channel: str | None = None
    result_channel: str | None = None

    @property
    def effective_command_channel(self) -> str:
        """實際使用的指令 channel"""
        return self.command_channel or f"channel:commands:{self.site_id}:write"

    @property
    def effective_result_channel(self) -> str:
        """實際使用的結果 channel"""
        return self.result_channel or f"channel:commands:{self.site_id}:result"


@dataclass(frozen=True)
class DistributedConfig:
    """
    分散式控制器配置

    Attributes:
        sites: 遠端站台配置列表
        trait_device_map: trait -> device_id 列表的映射
        poll_interval: 設備狀態輪詢間隔（秒）
        command_timeout: 指令執行超時（秒）
        system_alarm_on_device_offline: 設備離線時是否觸發 system_alarm
    """

    sites: list[RemoteSiteConfig] = field(default_factory=list)
    trait_device_map: dict[str, list[str]] = field(default_factory=dict)
    poll_interval: float = 1.0
    command_timeout: float = 5.0
    system_alarm_on_device_offline: bool = True

    @property
    def all_device_ids(self) -> list[str]:
        """所有站台的設備 ID 列表"""
        result: list[str] = []
        for site in self.sites:
            result.extend(site.device_ids)
        return result

    @property
    def device_site_map(self) -> dict[str, RemoteSiteConfig]:
        """device_id -> 所屬站台配置的映射"""
        result: dict[str, RemoteSiteConfig] = {}
        for site in self.sites:
            for device_id in site.device_ids:
                result[device_id] = site
        return result


__all__ = [
    "DistributedConfig",
    "RemoteSiteConfig",
]



# =============== CAN Client - python-can Implementation ===============
#
# 基於 python-can 的 CAN 客戶端實作
#
# 支援 SocketCAN、TCP 閘道器等多種介面。

from __future__ import annotations

import asyncio
import time
from typing import Any, Callable

from csp_lib.can.config import CANBusConfig, CANFrame
from csp_lib.can.exceptions import CANConnectionError, CANSendError, CANTimeoutError
from csp_lib.core import get_logger

from .base import AsyncCANClientBase

logger = get_logger(__name__)


class PythonCANClient(AsyncCANClientBase):
    """
    基於 python-can 的 CAN 客戶端

    將 python-can 的阻塞 I/O 包裝為 async 介面。
    背景 listener task 持續接收訊框，按 CAN ID 分發到已訂閱的 handlers。

    使用範例::

        client = PythonCANClient(CANBusConfig(interface="socketcan", channel="can0"))
        await client.connect()
        await client.start_listener()

        cancel = client.subscribe(0x100, lambda frame: print(frame))
        await client.send(0x200, b"\\x01\\x02")

        await client.stop_listener()
        await client.disconnect()
    """

    def __init__(self, config: CANBusConfig) -> None:
        self._config = config
        self._bus: Any = None  # can.Bus instance
        self._connected = False
        self._listener_task: asyncio.Task[None] | None = None
        self._handlers: dict[int, list[Callable[[CANFrame], Any]]] = {}
        self._pending_responses: dict[int, asyncio.Future[CANFrame]] = {}

    async def connect(self) -> None:
        if self._connected:
            return
        try:
            import can

            self._bus = await asyncio.to_thread(
                can.Bus,
                interface=self._config.interface,
                channel=self._config.channel,
                bitrate=self._config.bitrate,
                receive_own_messages=self._config.receive_own_messages,
            )
            self._connected = True
            logger.info(
                "CAN 連線成功: interface={}, channel={}",
                self._config.interface,
                self._config.channel,
            )
        except ImportError as e:
            raise CANConnectionError(
                "python-can 未安裝，請使用 pip install python-can 或 pip install csp0924_lib[can]"
            ) from e
        except Exception as e:
            raise CANConnectionError(f"CAN 連線失敗: {e}") from e

    async def disconnect(self) -> None:
        await self.stop_listener()
        if self._bus is not None:
            try:
                await asyncio.to_thread(self._bus.shutdown)
            except Exception as e:
                logger.warning("CAN 斷線錯誤: {}", e)
            self._bus = None
        self._connected = False

    async def is_connected(self) -> bool:
        return self._connected

    async def start_listener(self) -> None:
        if self._listener_task is not None:
            return
        self._listener_task = asyncio.create_task(self._listen_loop(), name="can_listener")

    async def stop_listener(self) -> None:
        if self._listener_task is None:
            return
        self._listener_task.cancel()
        try:
            await self._listener_task
        except asyncio.CancelledError:
            pass
        self._listener_task = None

    def subscribe(self, can_id: int, handler: Callable[[CANFrame], Any]) -> Callable[[], None]:
        if can_id not in self._handlers:
            self._handlers[can_id] = []
        self._handlers[can_id].append(handler)

        def cancel() -> None:
            if can_id in self._handlers and handler in self._handlers[can_id]:
                self._handlers[can_id].remove(handler)

        return cancel

    async def send(self, can_id: int, data: bytes) -> None:
        if not self._connected or self._bus is None:
            raise CANSendError("CAN 未連線")
        try:
            import can

            msg = can.Message(arbitration_id=can_id, data=data, is_extended_id=False)
            await asyncio.to_thread(self._bus.send, msg)
        except Exception as e:
            raise CANSendError(f"CAN 發送失敗: {e}") from e

    async def request(
        self,
        can_id: int,
        data: bytes,
        response_id: int,
        timeout: float = 1.0,
    ) -> CANFrame:
        loop = asyncio.get_running_loop()
        future: asyncio.Future[CANFrame] = loop.create_future()
        self._pending_responses[response_id] = future

        try:
            await self.send(can_id, data)
            return await asyncio.wait_for(future, timeout=timeout)
        except asyncio.TimeoutError as e:
            raise CANTimeoutError(f"等待 CAN ID 0x{response_id:03X} 回應逾時 ({timeout}s)") from e
        finally:
            self._pending_responses.pop(response_id, None)

    async def _listen_loop(self) -> None:
        """背景接收循環"""
        while True:
            try:
                msg = await asyncio.to_thread(self._bus.recv, 1.0)
                if msg is None:
                    continue

                frame = CANFrame(
                    can_id=msg.arbitration_id,
                    data=bytes(msg.data),
                    timestamp=msg.timestamp or time.time(),
                    is_remote=msg.is_remote_frame,
                )

                # 優先處理 request-response
                pending = self._pending_responses.get(frame.can_id)
                if pending is not None and not pending.done():
                    pending.set_result(frame)

                # 分發給訂閱者
                handlers = self._handlers.get(frame.can_id, [])
                for handler in handlers:
                    try:
                        result = handler(frame)
                        if asyncio.iscoroutine(result):
                            await result
                    except Exception:
                        logger.opt(exception=True).warning("CAN handler 錯誤: can_id=0x{:03X}", frame.can_id)

            except asyncio.CancelledError:
                break
            except Exception:
                logger.opt(exception=True).warning("CAN listener 錯誤")


__all__ = [
    "PythonCANClient",
]



"""Health API endpoint."""

from __future__ import annotations

from typing import Any

from fastapi import APIRouter, Depends

from csp_lib.core.health import HealthReport
from csp_lib.integration import SystemController

from ..dependencies import get_system_controller

router = APIRouter(tags=["health"])


def _serialize_health(report: HealthReport) -> dict[str, Any]:
    """Recursively serialize a HealthReport."""
    return {
        "status": report.status.value,
        "component": report.component,
        "message": report.message,
        "details": report.details,
        "children": [_serialize_health(c) for c in report.children],
    }


@router.get("/health")
def get_health(sc: SystemController = Depends(get_system_controller)) -> dict[str, Any]:
    """Get system health report."""
    report = sc.health()
    return _serialize_health(report)



# =============== Power Compensator ===============
#
# 前饋 + 積分閉環功率補償器
#
# 控制方程式：command = ff(power_bin) × setpoint + Ki × ∫error·dt
#
# 職責：
#   - 前饋補償：按功率區間查表，補償 PCS 非線性與輔電損耗
#   - 積分修正：短期殘差修正，含 anti-windup / deadband / clamp
#   - 穩態學習：I 項貢獻吸收進 FF 表，長期自適應
#   - 暫態閘門：setpoint 變更後等 PCS 到位才啟動 I
#   - FF 表持久化：JSON 檔案存/讀
#
# 實作 CommandProcessor Protocol，可作為 post_protection_processor 使用。
#
# Usage::
#
#     compensator = PowerCompensator(PowerCompensatorConfig(
#         rated_power=2000.0,
#         measurement_key="meter_power",
#     ))
#
#     config = SystemControllerConfig(
#         post_protection_processors=[compensator],
#     )

from __future__ import annotations

import asyncio
import json
import math
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Protocol, runtime_checkable

from csp_lib.controller.core import Command, StrategyContext
from csp_lib.core import get_logger

logger = get_logger(__name__)


# =============== FF Table Repository ===============


@runtime_checkable
class FFTableRepository(Protocol):
    """
    FF Table 持久化介面

    實作此 Protocol 即可支援任意後端（JSON、MongoDB、Redis 等）。
    PowerCompensator 透過此介面存取 FF Table。
    """

    def save(self, table: dict[int, float]) -> None:
        """儲存 FF Table"""
        ...

    def load(self) -> dict[int, float] | None:
        """載入 FF Table，不存在時回傳 None"""
        ...


class JsonFFTableRepository:
    """
    JSON 檔案持久化（預設實作）

    支援不同 step 的 FF Table 自動遷移（線性插值）。

    Args:
        path: JSON 檔案路徑
        power_bin_step_pct: 當前 FF Table 的 step 百分比（用於遷移偵測）
    """

    def __init__(self, path: str, power_bin_step_pct: int = 5) -> None:
        self._path = path
        self._step_pct = power_bin_step_pct

    def save(self, table: dict[int, float]) -> None:
        try:
            data = {str(k): v for k, v in table.items()}
            Path(self._path).parent.mkdir(parents=True, exist_ok=True)
            with open(self._path, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2)
        except Exception:
            logger.opt(exception=True).warning("JsonFFTableRepository: failed to save")

    def load(self) -> dict[int, float] | None:
        if not Path(self._path).exists():
            return None
        try:
            with open(self._path, encoding="utf-8") as f:
                data = json.load(f)
            if not data:
                return None
            return {int(k): float(v) for k, v in data.items()}
        except Exception:
            logger.opt(exception=True).warning("JsonFFTableRepository: failed to load")
            return None


class MongoFFTableRepository:
    """
    MongoDB 持久化

    將 FF Table 存為單一 document，格式::

        {
            "_id": <document_id>,
            "table": {"0": 1.0, "5": 1.02, "-5": 0.98, ...},
            "updated_at": <datetime>
        }

    Args:
        collection: motor AsyncIOMotorCollection 實例
        document_id: document 的 _id（預設 "ff_table"）
    """

    def __init__(self, collection: Any, document_id: str = "ff_table") -> None:
        self._collection = collection
        self._doc_id = document_id

    def save(self, table: dict[int, float]) -> None:
        """同步包裝 — 在 asyncio event loop 中排程 async save"""

        try:
            loop = asyncio.get_running_loop()
            loop.create_task(self._async_save(table))
        except RuntimeError:
            logger.warning("MongoFFTableRepository: no running event loop, skip save")

    def load(self) -> dict[int, float] | None:
        """同步包裝 — 用於 PowerCompensator.__init__，無法 await"""
        # MongoDB load 需要 async，init 時無法 await
        # 由 async_load() 在 start 階段呼叫
        logger.debug("MongoFFTableRepository: sync load not available, use async_load()")
        return None

    async def _async_save(self, table: dict[int, float]) -> None:
        from datetime import datetime, timezone

        try:
            data = {str(k): v for k, v in table.items()}
            await self._collection.update_one(
                {"_id": self._doc_id},
                {"$set": {"table": data, "updated_at": datetime.now(timezone.utc)}},
                upsert=True,
            )
            logger.debug(f"MongoFFTableRepository: saved {len(table)} bins")
        except Exception:
            logger.opt(exception=True).warning("MongoFFTableRepository: failed to save")

    async def async_load(self) -> dict[int, float] | None:
        """
        Async 載入 FF Table

        在 PowerCompensator 啟動後呼叫::

            repo = MongoFFTableRepository(collection)
            table = await repo.async_load()
            if table:
                compensator.load_ff_table(table)
        """
        try:
            doc = await self._collection.find_one({"_id": self._doc_id})
            if doc and "table" in doc:
                return {int(k): float(v) for k, v in doc["table"].items()}
            return None
        except Exception:
            logger.opt(exception=True).warning("MongoFFTableRepository: failed to load")
            return None


@dataclass
class PowerCompensatorConfig:
    """
    功率補償器配置

    Attributes:
        rated_power: 系統額定功率 (kW)
        output_min: 輸出下限 (kW)
        output_max: 輸出上限 (kW)
        ki: 積分增益 (1/s)
        integral_max_ratio: I 項最大貢獻 = ratio × rated_power
        deadband: 死區 (kW)，誤差低於此值不累積 I
        power_bin_step_pct: FF 表功率區間寬度 (% of rated)
        steady_state_threshold: 穩態門檻 |error/setpoint|
        steady_state_cycles: 連續穩態週期數，達標後觸發 FF 學習
        settle_ratio: 暫態閘門比例，setpoint 變更後 |error| > 變化量 × settle_ratio 時不倒數
        hold_cycles: setpoint 變更後暫停積分的週期數
        ff_min: FF 補償係數下限
        ff_max: FF 補償係數上限
        error_ema_alpha: 誤差 EMA 濾波係數 (0=停用)
        rate_limit: 輸出變化率限制 (kW/s, 0=停用)
        measurement_key: context.extra 中量測值的 key
        persist_path: FF 表持久化路徑 (空=不持久化)
    """

    rated_power: float = 2000.0
    output_min: float = -2000.0
    output_max: float = 2000.0
    ki: float = 0.3
    integral_max_ratio: float = 0.05
    deadband: float = 0.5
    power_bin_step_pct: int = 5
    steady_state_threshold: float = 0.02
    steady_state_cycles: int = 5
    settle_ratio: float = 0.15
    hold_cycles: int = 2
    ff_min: float = 0.8
    ff_max: float = 1.5
    error_ema_alpha: float = 0.0
    rate_limit: float = 0.0
    measurement_key: str = "meter_power"
    persist_path: str = ""


class PowerCompensator:
    """
    前饋 + 積分閉環功率補償器

    實作 CommandProcessor Protocol，可在 SystemControllerConfig 的
    post_protection_processors 中使用。

    command = ff(power_bin) × setpoint + Ki × ∫error·dt

    正功率（放電）：ff_output = ff × setpoint
    負功率（充電）：ff_output = setpoint / ff（補償輔電使電網讀數更負的效果）

    Usage::

        compensator = PowerCompensator(PowerCompensatorConfig(
            rated_power=2000.0,
            measurement_key="meter_power",
        ))

        # 作為 CommandProcessor
        config = SystemControllerConfig(
            post_protection_processors=[compensator],
        )

        # 或手動呼叫
        compensated_cmd = await compensator.process(command, context)
    """

    def __init__(
        self,
        config: PowerCompensatorConfig | None = None,
        repository: FFTableRepository | None = None,
    ) -> None:
        self._config = config or PowerCompensatorConfig()
        self._enabled = True

        # Repository：優先使用注入的，否則從 config.persist_path 建立 JSON repo
        if repository is not None:
            self._repository: FFTableRepository | None = repository
        elif self._config.persist_path:
            self._repository = JsonFFTableRepository(self._config.persist_path, self._config.power_bin_step_pct)
        else:
            self._repository = None

        # FF 補償表：bin_index → ff_factor (初始 1.0)
        self._ff_table: dict[int, float] = self._build_initial_ff_table()
        self._load_ff_table()

        # 積分與濾波狀態
        self._integral: float = 0.0
        self._filtered_error: float = 0.0

        # 追蹤狀態
        self._last_setpoint: float = 0.0
        self._last_output: float = 0.0
        self._steady_count: int = 0
        self._integral_hold: int = 0
        self._settle_threshold: float = 0.0

    # ─────────────────────── CommandProcessor Protocol ───────────────────────

    async def process(self, command: Command, context: StrategyContext) -> Command:
        """
        CommandProcessor.process() 實作

        從 context.extra 讀取量測值，計算補償後的功率指令。
        若補償器停用或無量測值，直接回傳原始 command。
        """
        if not self._enabled:
            return command

        measurement_key = self._config.measurement_key
        measurement = context.extra.get(measurement_key)
        if measurement is None:
            return command

        # 計算 dt（從 context.extra 取得，或預設 0.3s）
        dt = float(context.extra.get("dt", 0.3))

        compensated_p = self.compensate(
            setpoint=command.p_target,
            measurement=float(measurement),
            dt=dt,
        )
        return command.with_p(compensated_p)

    # ─────────────────────── 核心演算法 ───────────────────────

    def compensate(self, setpoint: float, measurement: float, dt: float) -> float:
        """
        計算補償後的 PCS 功率指令

        Args:
            setpoint: 目標功率 (kW)，保護鏈輸出
            measurement: 電網端實際功率 (kW)
            dt: 距上次呼叫的時間間隔 (秒)

        Returns:
            補償後的 PCS 功率指令 (kW)
        """
        cfg = self._config

        # 零目標 → 直接輸出 0，重置積分
        if abs(setpoint) < 1e-6:
            self._integral = 0.0
            self._last_setpoint = 0.0
            self._last_output = 0.0
            self._steady_count = 0
            return 0.0

        # 1. Setpoint 變動策略
        self._apply_setpoint_change_policy(setpoint)
        self._last_setpoint = setpoint

        # 2. 計算誤差
        error = setpoint - measurement

        # 3. 誤差濾波
        if cfg.error_ema_alpha > 0:
            self._filtered_error = cfg.error_ema_alpha * error + (1.0 - cfg.error_ema_alpha) * self._filtered_error
            filtered_error = self._filtered_error
        else:
            filtered_error = error

        # 4. 前饋查表
        ff = self._get_ff(setpoint)
        if setpoint >= 0:
            ff_output = ff * setpoint
        else:
            ff_output = setpoint / ff if abs(ff) > 1e-6 else setpoint

        # 5. 飽和檢測
        saturated = ff_output >= cfg.output_max or ff_output <= cfg.output_min

        # 6. 積分更新
        if saturated:
            self._integral = 0.0
        elif self._integral_hold > 0:
            if abs(filtered_error) <= self._settle_threshold:
                self._integral_hold -= 1
        elif abs(filtered_error) >= cfg.deadband:
            self._integral += filtered_error * dt
            self._clamp_integral()

        # 7. 計算輸出
        output = ff_output + cfg.ki * self._integral

        # 8. 輸出限幅
        output = max(cfg.output_min, min(output, cfg.output_max))

        # 9. 變化率限制
        if cfg.rate_limit > 0:
            max_delta = cfg.rate_limit * dt
            delta = output - self._last_output
            if abs(delta) > max_delta:
                output = self._last_output + math.copysign(max_delta, delta)

        # 10. 穩態學習
        if not saturated:
            self._learn_if_steady(setpoint, filtered_error)

        self._last_output = output
        return output

    # ─────────────────────── 公開介面 ───────────────────────

    def reset(self) -> None:
        """重置積分與追蹤狀態（保留 FF 補償表）"""
        self._integral = 0.0
        self._filtered_error = 0.0
        self._last_setpoint = 0.0
        self._last_output = 0.0
        self._steady_count = 0
        self._integral_hold = 0

    def reset_ff_table(self) -> None:
        """重置 FF 補償表為全 1.0 並清除持久化"""
        self._ff_table = self._build_initial_ff_table()
        # 清除持久化
        if self._config.persist_path:
            try:
                Path(self._config.persist_path).unlink(missing_ok=True)
            except OSError:
                pass
        logger.info("FF table reset to defaults")

    def load_ff_table(self, table: dict[int, float]) -> None:
        """
        直接載入 FF Table（供外部校準使用）

        Args:
            table: {bin_index: ff_factor} 映射
        """
        loaded = 0
        for idx, ff in table.items():
            if idx in self._ff_table:
                self._ff_table[idx] = ff
                loaded += 1
        logger.info(f"FF table loaded externally ({loaded} bins)")

    async def async_init(self) -> None:
        """
        Async 初始化 — 從 async repository（如 MongoDB）載入 FF Table

        JSON repository 在 __init__ 中同步載入，不需要呼叫此方法。
        MongoDB repository 必須在 event loop 啟動後呼叫此方法。

        SystemController._on_start() 會自動呼叫所有 CommandProcessor 的 async_init()。

        Usage::

            comp = PowerCompensator(config, repository=MongoFFTableRepository(collection))
            await comp.async_init()  # 從 MongoDB 載入 FF Table
        """
        if self._repository is None:
            return
        if hasattr(self._repository, "async_load"):
            table = await self._repository.async_load()
            if table:
                self.load_ff_table(table)

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._enabled = value
        if not value:
            self.reset()

    @property
    def diagnostics(self) -> dict:
        """診斷資訊（供 logging / API 使用）"""
        cfg = self._config
        return {
            "enabled": self._enabled,
            "integral": round(self._integral, 4),
            "i_contribution": round(cfg.ki * self._integral, 2),
            "last_setpoint": round(self._last_setpoint, 1),
            "last_output": round(self._last_output, 1),
            "last_ff": round(self._get_ff(self._last_setpoint), 4),
            "steady_count": self._steady_count,
            "hold_remaining": self._integral_hold,
        }

    @property
    def ff_table(self) -> dict[int, float]:
        """FF 補償表的淺拷貝"""
        return dict(self._ff_table)

    # ─────────────────────── 內部方法 ───────────────────────

    def _build_initial_ff_table(self) -> dict[int, float]:
        step = self._config.power_bin_step_pct
        if step <= 0:
            step = 10
        n_bins = 100 // step
        return {i: 1.0 for i in range(-n_bins, n_bins + 1)}

    def _get_bin_index(self, power_kw: float) -> int:
        cfg = self._config
        pct = power_kw / cfg.rated_power * 100.0
        idx = round(pct / cfg.power_bin_step_pct)
        n_bins = 100 // cfg.power_bin_step_pct
        return max(-n_bins, min(idx, n_bins))

    def _get_ff(self, setpoint: float) -> float:
        idx = self._get_bin_index(setpoint)
        return self._ff_table.get(idx, 1.0)

    def _apply_setpoint_change_policy(self, new_setpoint: float) -> None:
        old = self._last_setpoint
        if abs(new_setpoint - old) < 0.1:
            return
        self._integral = 0.0
        self._steady_count = 0
        self._integral_hold = self._config.hold_cycles
        self._settle_threshold = abs(new_setpoint - old) * self._config.settle_ratio
        self._inherit_ff(old, new_setpoint)

    def _inherit_ff(self, old_setpoint: float, new_setpoint: float) -> None:
        old_idx = self._get_bin_index(old_setpoint)
        new_idx = self._get_bin_index(new_setpoint)
        if old_idx == new_idx:
            return
        old_ff = self._ff_table.get(old_idx, 1.0)
        new_ff = self._ff_table.get(new_idx, 1.0)
        if abs(new_ff - 1.0) < 1e-6 and abs(old_ff - 1.0) > 1e-6:
            self._ff_table[new_idx] = old_ff
            logger.debug(f"FF inherit: bin[{old_idx}] → bin[{new_idx}], ff={old_ff:.4f}")

    def _clamp_integral(self) -> None:
        cfg = self._config
        if cfg.ki == 0:
            return
        max_contribution = cfg.integral_max_ratio * cfg.rated_power
        integral_max = max_contribution / cfg.ki
        self._integral = max(-integral_max, min(self._integral, integral_max))

    def _learn_if_steady(self, setpoint: float, filtered_error: float) -> None:
        cfg = self._config
        if abs(setpoint) < cfg.deadband:
            self._steady_count = 0
            return

        relative_error = abs(filtered_error / setpoint)
        if relative_error < cfg.steady_state_threshold:
            self._steady_count += 1
        else:
            self._steady_count = 0
            return

        if self._steady_count < cfg.steady_state_cycles:
            return

        i_term = cfg.ki * self._integral
        if abs(i_term) < cfg.deadband:
            self._steady_count = 0
            return

        idx = self._get_bin_index(setpoint)
        old_ff = self._ff_table.get(idx, 1.0)
        if setpoint >= 0:
            new_ff = (old_ff * setpoint + i_term) / setpoint
        else:
            denom = setpoint / old_ff + i_term
            new_ff = setpoint / denom if abs(denom) > 1e-6 else old_ff

        new_ff = max(cfg.ff_min, min(new_ff, cfg.ff_max))

        if abs(new_ff - old_ff) > 1e-6:
            self._ff_table[idx] = new_ff
            self._integral = 0.0
            logger.debug(f"FF learn: bin[{idx}]({idx * cfg.power_bin_step_pct}%) ff: {old_ff:.4f} → {new_ff:.4f}")
            self._save_ff_table()

        self._steady_count = 0

    # ─────────────────────── 持久化 ───────────────────────

    def _save_ff_table(self) -> None:
        """透過 repository 儲存 FF Table"""
        if self._repository is None:
            return
        self._repository.save(dict(self._ff_table))

    def _load_ff_table(self) -> None:
        """透過 repository 載入 FF Table（含 step 遷移）"""
        if self._repository is None:
            return
        data = self._repository.load()
        if not data:
            return

        file_max_idx = max(abs(i) for i in data.keys()) if data else 0
        cur_n_bins = 100 // self._config.power_bin_step_pct

        if file_max_idx != cur_n_bins and file_max_idx > 0:
            # step 不同 → 線性插值遷移
            old_step_pct = 100 // file_max_idx
            ratio = old_step_pct / self._config.power_bin_step_pct
            migrated = 0
            for new_idx in self._ff_table:
                old_pos = new_idx / ratio
                lo = math.floor(old_pos)
                hi = math.ceil(old_pos)
                if lo == hi:
                    if lo in data:
                        self._ff_table[new_idx] = data[lo]
                        migrated += 1
                elif lo in data and hi in data:
                    frac = old_pos - lo
                    self._ff_table[new_idx] = data[lo] * (1 - frac) + data[hi] * frac
                    migrated += 1
            logger.info(
                f"FF table migrated: old step={old_step_pct}% → new step={self._config.power_bin_step_pct}%, "
                f"interpolated {migrated} bins"
            )
            self._save_ff_table()
        else:
            loaded = 0
            for idx, ff in data.items():
                if idx in self._ff_table:
                    self._ff_table[idx] = ff
                    loaded += 1
            logger.info(f"FF table loaded ({loaded} bins)")


__all__ = [
    "FFTableRepository",
    "JsonFFTableRepository",
    "MongoFFTableRepository",
    "PowerCompensator",
    "PowerCompensatorConfig",
]



# =============== Monitor - Alarm ===============
#
# 系統告警評估
#
# 複用 equipment.alarm 模組的 ThresholdAlarmEvaluator 與 AlarmStateManager，
# 提供系統指標（CPU/RAM/Disk）的閾值告警：
#   - SystemAlarmEvaluator: 系統告警評估外觀

from __future__ import annotations

from typing import TYPE_CHECKING

from csp_lib.equipment.alarm.definition import AlarmDefinition, AlarmLevel, HysteresisConfig
from csp_lib.equipment.alarm.evaluator import Operator, ThresholdAlarmEvaluator, ThresholdCondition
from csp_lib.equipment.alarm.state import AlarmEvent, AlarmStateManager

if TYPE_CHECKING:
    from csp_lib.monitor.collector import SystemMetrics
    from csp_lib.monitor.config import MonitorConfig


def create_system_alarm_evaluators(config: MonitorConfig) -> dict[str, ThresholdAlarmEvaluator]:
    """
    建立系統告警評估器

    Args:
        config: 監控器配置

    Returns:
        {指標名稱: 閾值評估器} 字典
    """
    hysteresis = HysteresisConfig(
        activate_threshold=config.hysteresis_activate,
        clear_threshold=config.hysteresis_clear,
    )
    evaluators: dict[str, ThresholdAlarmEvaluator] = {}

    if config.enable_cpu:
        evaluators["cpu_percent"] = ThresholdAlarmEvaluator(
            point_name="cpu_percent",
            conditions=[
                ThresholdCondition(
                    alarm=AlarmDefinition(
                        code="SYS_CPU_HIGH",
                        name="系統 CPU 使用率過高",
                        level=AlarmLevel.WARNING,
                        hysteresis=hysteresis,
                        description=f"CPU 使用率超過 {config.thresholds.cpu_percent}%",
                    ),
                    operator=Operator.GT,
                    value=config.thresholds.cpu_percent,
                ),
            ],
        )

    if config.enable_ram:
        evaluators["ram_percent"] = ThresholdAlarmEvaluator(
            point_name="ram_percent",
            conditions=[
                ThresholdCondition(
                    alarm=AlarmDefinition(
                        code="SYS_RAM_HIGH",
                        name="系統 RAM 使用率過高",
                        level=AlarmLevel.WARNING,
                        hysteresis=hysteresis,
                        description=f"RAM 使用率超過 {config.thresholds.ram_percent}%",
                    ),
                    operator=Operator.GT,
                    value=config.thresholds.ram_percent,
                ),
            ],
        )

    if config.enable_disk:
        for path in config.disk_paths:
            safe_path = path.replace("/", "_").strip("_") or "root"
            code = f"SYS_DISK_HIGH_{safe_path}"
            evaluators[f"disk:{path}"] = ThresholdAlarmEvaluator(
                point_name=f"disk:{path}",
                conditions=[
                    ThresholdCondition(
                        alarm=AlarmDefinition(
                            code=code,
                            name=f"磁碟 {path} 使用率過高",
                            level=AlarmLevel.ALARM,
                            hysteresis=hysteresis,
                            description=f"磁碟 {path} 使用率超過 {config.thresholds.disk_percent}%",
                        ),
                        operator=Operator.GT,
                        value=config.thresholds.disk_percent,
                    ),
                ],
            )

    return evaluators


class SystemAlarmEvaluator:
    """
    系統告警評估外觀

    整合 ThresholdAlarmEvaluator 與 AlarmStateManager，
    提供單一 evaluate(metrics) 介面。
    """

    def __init__(self, config: MonitorConfig) -> None:
        self._config = config
        self._evaluators = create_system_alarm_evaluators(config)
        self._state_manager = AlarmStateManager()
        self._network_evaluators: dict[str, dict[str, ThresholdAlarmEvaluator]] = {}

        # 註冊所有告警定義
        for evaluator in self._evaluators.values():
            self._state_manager.register_alarms(evaluator.get_alarms())

    def _get_or_create_network_evaluators(self, iface_name: str) -> dict[str, ThresholdAlarmEvaluator]:
        """取得或建立網路介面告警評估器"""
        if iface_name in self._network_evaluators:
            return self._network_evaluators[iface_name]

        thresholds = self._config.network_thresholds
        hysteresis = HysteresisConfig(
            activate_threshold=self._config.hysteresis_activate,
            clear_threshold=self._config.hysteresis_clear,
        )
        safe_name = iface_name.replace("/", "_").replace("\\", "_").replace(":", "_")
        evaluators: dict[str, ThresholdAlarmEvaluator] = {}

        if thresholds.send_rate_bytes > 0:
            code = f"SYS_NET_SEND_HIGH_{safe_name}"
            ev = ThresholdAlarmEvaluator(
                point_name=f"net_send:{iface_name}",
                conditions=[
                    ThresholdCondition(
                        alarm=AlarmDefinition(
                            code=code,
                            name=f"網路介面 {iface_name} 發送速率過高",
                            level=AlarmLevel.WARNING,
                            hysteresis=hysteresis,
                            description=f"網路介面 {iface_name} 發送速率超過 {thresholds.send_rate_bytes} bytes/s",
                        ),
                        operator=Operator.GT,
                        value=thresholds.send_rate_bytes,
                    ),
                ],
            )
            evaluators["send"] = ev
            self._state_manager.register_alarms(ev.get_alarms())

        if thresholds.recv_rate_bytes > 0:
            code = f"SYS_NET_RECV_HIGH_{safe_name}"
            ev = ThresholdAlarmEvaluator(
                point_name=f"net_recv:{iface_name}",
                conditions=[
                    ThresholdCondition(
                        alarm=AlarmDefinition(
                            code=code,
                            name=f"網路介面 {iface_name} 接收速率過高",
                            level=AlarmLevel.WARNING,
                            hysteresis=hysteresis,
                            description=f"網路介面 {iface_name} 接收速率超過 {thresholds.recv_rate_bytes} bytes/s",
                        ),
                        operator=Operator.GT,
                        value=thresholds.recv_rate_bytes,
                    ),
                ],
            )
            evaluators["recv"] = ev
            self._state_manager.register_alarms(ev.get_alarms())

        self._network_evaluators[iface_name] = evaluators
        return evaluators

    def evaluate(self, metrics: SystemMetrics) -> list[AlarmEvent]:
        """
        評估系統指標並返回告警事件

        Args:
            metrics: 系統指標

        Returns:
            告警事件列表（狀態變化時產生）
        """
        all_evaluations: dict[str, bool] = {}

        # CPU
        if "cpu_percent" in self._evaluators:
            result = self._evaluators["cpu_percent"].evaluate(metrics.cpu_percent)
            all_evaluations.update(result)

        # RAM
        if "ram_percent" in self._evaluators:
            result = self._evaluators["ram_percent"].evaluate(metrics.ram_percent)
            all_evaluations.update(result)

        # Disk
        for path, percent in metrics.disk_usage.items():
            key = f"disk:{path}"
            if key in self._evaluators:
                result = self._evaluators[key].evaluate(percent)
                all_evaluations.update(result)

        # Network interfaces
        if self._config.network_thresholds.is_enabled:
            for iface_name, iface_metrics in metrics.interface_metrics.items():
                net_evaluators = self._get_or_create_network_evaluators(iface_name)
                if "send" in net_evaluators:
                    result = net_evaluators["send"].evaluate(iface_metrics.send_rate)
                    all_evaluations.update(result)
                if "recv" in net_evaluators:
                    result = net_evaluators["recv"].evaluate(iface_metrics.recv_rate)
                    all_evaluations.update(result)

        return self._state_manager.update(all_evaluations)

    @property
    def active_alarms(self) -> list[str]:
        """取得活躍告警代碼列表"""
        return [state.definition.code for state in self._state_manager.get_active_alarms()]

    @property
    def state_manager(self) -> AlarmStateManager:
        """取得告警狀態管理器"""
        return self._state_manager


__all__ = [
    "SystemAlarmEvaluator",
    "create_system_alarm_evaluators",
]



# =============== Notification - Config ===============
#
# 通知配置
#
# 定義通知分發相關參數：
#   - NotificationConfig: 事件標籤與標題模板

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class NotificationConfig:
    """
    通知配置

    Attributes:
        triggered_label: 告警觸發時的事件標籤
        resolved_label: 告警解除時的事件標籤
        title_template: 通知標題模板，支援 {level}, {device_id}, {name}, {event_label} 佔位符
    """

    triggered_label: str = "觸發"
    resolved_label: str = "解除"
    title_template: str = "[{level}] {device_id} {name} - {event_label}"

    def __post_init__(self) -> None:
        if not self.triggered_label:
            raise ValueError("triggered_label 不可為空")
        if not self.resolved_label:
            raise ValueError("resolved_label 不可為空")
        if not self.title_template:
            raise ValueError("title_template 不可為空")


__all__ = [
    "NotificationConfig",
]



# =============== Integration - Registry ===============
#
# Trait-based 設備查詢索引
#
# 維護 device_id ↔ trait 的雙向索引：
#   - 依 device_id 查詢設備
#   - 依 trait 查詢所有匹配設備（支援 responsive 過濾）
#   - 依 capability 查詢具備指定能力的設備
#   - validate_capabilities: 驗證能力需求是否滿足（preflight check）
#   - 不管理設備生命週期，僅做查詢索引

from __future__ import annotations

import threading
from typing import TYPE_CHECKING, Any, Sequence

from csp_lib.core import get_logger

from .schema import CapabilityRequirement, capability_display_name

if TYPE_CHECKING:
    from csp_lib.equipment.device import AsyncModbusDevice
    from csp_lib.equipment.device.capability import Capability

logger = get_logger(__name__)


class DeviceRegistry:
    """
    Trait-based 設備查詢索引

    維護雙向索引：
      - device_id → AsyncModbusDevice
      - device_id → set[trait]
      - trait → set[device_id]

    不管理設備生命週期，僅負責查詢。
    所有依 trait 查詢的結果皆按 device_id 排序，確保確定性。
    """

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._devices: dict[str, AsyncModbusDevice] = {}  # device_id → 設備實例
        self._device_traits: dict[str, set[str]] = {}  # device_id → 該設備的 traits
        self._trait_devices: dict[str, set[str]] = {}  # trait → 擁有該 trait 的 device_ids
        self._metadata: dict[str, dict[str, Any]] = {}  # device_id → 靜態 metadata

    # ---- 註冊 / 移除 ----

    def register(
        self,
        device: AsyncModbusDevice,
        traits: list[str] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """
        註冊設備與可選的 traits 和 metadata

        Args:
            device: 要註冊的 Modbus 設備
            traits: 設備的 trait 標籤列表（可選）
            metadata: 設備靜態資訊（可選），如 rated_p、rated_s 等

        Raises:
            ValueError: device_id 已存在時拋出，防止靜默覆蓋
        """
        did = device.device_id
        with self._lock:
            if did in self._devices:
                raise ValueError(f"Device '{did}' is already registered.")
            self._devices[did] = device
            self._device_traits[did] = set()
            self._metadata[did] = dict(metadata) if metadata else {}
            for trait in traits or []:
                self._add_trait_index(did, trait)

    def register_with_capabilities(
        self,
        device: AsyncModbusDevice,
        extra_traits: Sequence[str] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """
        Register device with auto-discovered traits from capabilities.

        Auto-generates traits like "cap:active_power_control" from device capabilities,
        then merges with extra_traits provided by user.

        Args:
            device: 要註冊的 Modbus 設備
            extra_traits: 額外的 trait 標籤（可選），會排在自動發現的 traits 前面
            metadata: 設備靜態資訊（可選）

        Raises:
            ValueError: device_id 已存在時拋出
        """
        auto_traits = [f"cap:{cap_name}" for cap_name in device.capabilities]
        all_traits = list(extra_traits or []) + auto_traits
        self.register(device, traits=all_traits, metadata=metadata or {})

    def unregister(self, device_id: str) -> None:
        """
        移除設備及其所有 trait 關聯

        Args:
            device_id: 要移除的設備 ID（不存在時靜默忽略）
        """
        with self._lock:
            if device_id not in self._devices:
                return
            for trait in list(self._device_traits.get(device_id, [])):
                self._remove_trait_index(device_id, trait)
            del self._devices[device_id]
            del self._device_traits[device_id]
            self._metadata.pop(device_id, None)

    # ---- Trait 管理 ----

    def add_trait(self, device_id: str, trait: str) -> None:
        """
        為已註冊設備新增 trait

        Args:
            device_id: 目標設備 ID
            trait: 要新增的 trait 標籤

        Raises:
            KeyError: device_id 未註冊時拋出
        """
        with self._lock:
            if device_id not in self._devices:
                raise KeyError(f"Device '{device_id}' is not registered.")
            self._add_trait_index(device_id, trait)

    def remove_trait(self, device_id: str, trait: str) -> None:
        """
        移除設備的 trait

        Args:
            device_id: 目標設備 ID
            trait: 要移除的 trait 標籤

        Raises:
            KeyError: device_id 未註冊時拋出
        """
        with self._lock:
            if device_id not in self._devices:
                raise KeyError(f"Device '{device_id}' is not registered.")
            self._remove_trait_index(device_id, trait)

    # ---- 查詢 ----

    def get_device(self, device_id: str) -> AsyncModbusDevice | None:
        """依 ID 查詢設備，不存在回傳 None"""
        with self._lock:
            return self._devices.get(device_id)

    def get_devices_by_trait(self, trait: str) -> list[AsyncModbusDevice]:
        """依 trait 查詢所有設備（按 device_id 排序，確保確定性）"""
        with self._lock:
            ids = self._trait_devices.get(trait, set())
            return [self._devices[did] for did in sorted(ids)]

    def get_responsive_devices_by_trait(self, trait: str) -> list[AsyncModbusDevice]:
        """依 trait 查詢所有 is_responsive=True 的設備（按 device_id 排序）"""
        return [d for d in self.get_devices_by_trait(trait) if d.is_responsive]

    def get_first_responsive_device_by_trait(self, trait: str) -> AsyncModbusDevice | None:
        """依 trait 取得第一台 responsive 設備，無則回傳 None"""
        devices = self.get_responsive_devices_by_trait(trait)
        return devices[0] if devices else None

    def get_traits(self, device_id: str) -> set[str]:
        """取得設備的所有 traits，未註冊時回傳空集合"""
        with self._lock:
            return set(self._device_traits.get(device_id, set()))

    def get_metadata(self, device_id: str) -> dict[str, Any]:
        """取得設備的靜態 metadata，未註冊時回傳空 dict"""
        with self._lock:
            return dict(self._metadata.get(device_id, {}))

    @property
    def all_devices(self) -> list[AsyncModbusDevice]:
        """所有已註冊設備（按 device_id 排序）"""
        with self._lock:
            return [self._devices[did] for did in sorted(self._devices)]

    @property
    def all_traits(self) -> list[str]:
        """所有已知的 trait 標籤（排序）"""
        with self._lock:
            return sorted(self._trait_devices.keys())

    # ---- Capability 查詢 ----

    def get_devices_with_capability(self, capability: Capability | str) -> list[AsyncModbusDevice]:
        """取得具備指定能力的所有設備（按 device_id 排序）"""
        with self._lock:
            return sorted(
                [d for d in self._devices.values() if d.has_capability(capability)],
                key=lambda d: d.device_id,
            )

    def get_responsive_devices_with_capability(self, capability: Capability | str) -> list[AsyncModbusDevice]:
        """取得具備指定能力且 responsive 的設備（按 device_id 排序）"""
        return [d for d in self.get_devices_with_capability(capability) if d.is_responsive]

    def validate_capabilities(self, requirements: list[CapabilityRequirement]) -> list[str]:
        """驗證設備能力是否滿足需求

        Args:
            requirements: 能力需求列表

        Returns:
            不滿足的需求描述列表（空 = 全部通過）
        """
        failures: list[str] = []
        for req in requirements:
            if req.trait_filter:
                devices = [d for d in self.get_devices_by_trait(req.trait_filter) if d.has_capability(req.capability)]
            else:
                devices = self.get_devices_with_capability(req.capability)

            if len(devices) < req.min_count:
                cap_name = capability_display_name(req.capability)
                trait_suffix = f" with trait {req.trait_filter!r}" if req.trait_filter else ""
                failures.append(
                    f"Capability '{cap_name}' requires {req.min_count} device(s){trait_suffix}, found {len(devices)}"
                )
        return failures

    def get_capability_map(self) -> dict[str, list[str]]:
        """取得 capability_name → device_ids 映射

        Returns:
            capability 名稱對應具備該能力的設備 ID 列表（按 device_id 排序）
        """
        with self._lock:
            result: dict[str, list[str]] = {}
            for did, device in self._devices.items():
                for cap_name in device.capabilities:
                    if cap_name not in result:
                        result[cap_name] = []
                    result[cap_name].append(did)
            # 排序 device_ids 確保確定性
            for cap_name in result:
                result[cap_name].sort()
            return result

    def get_capability_map_text(self) -> str:
        """取得格式化的 capability 映射文字表格

        Returns:
            格式化文字，每行一個 capability 及其設備列表
        """
        cap_map = self.get_capability_map()
        if not cap_map:
            return ""
        lines: list[str] = []
        for cap_name in sorted(cap_map):
            device_ids = cap_map[cap_name]
            count = len(device_ids)
            devices_str = ", ".join(device_ids)
            lines.append(f"{cap_name} ({count} devices): {devices_str}")
        return "\n".join(lines)

    def capability_health(self, capability: Capability | str) -> dict[str, Any]:
        """取得指定 capability 的健康狀態

        Args:
            capability: 能力定義或名稱

        Returns:
            包含 capability、total_devices、responsive_devices、
            responsive_ratio、devices 的字典
        """
        cap_name = capability.name if hasattr(capability, "name") else str(capability)
        devices = self.get_devices_with_capability(capability)
        total = len(devices)
        device_details: list[dict[str, Any]] = []
        responsive_count = 0
        for d in devices:
            is_resp = d.is_responsive
            if is_resp:
                responsive_count += 1
            device_details.append({"device_id": d.device_id, "is_responsive": is_resp})
        return {
            "capability": cap_name,
            "total_devices": total,
            "responsive_devices": responsive_count,
            "responsive_ratio": responsive_count / total if total > 0 else 0.0,
            "devices": device_details,
        }

    def refresh_capability_traits(self, device_id: str) -> None:
        """重新同步設備的 capability traits（cap:xxx）

        根據設備目前的 capabilities 重新建立 cap: 前綴的 traits，
        移除不再存在的、新增新出現的。

        Args:
            device_id: 目標設備 ID

        Raises:
            KeyError: device_id 未註冊時拋出
        """
        with self._lock:
            device = self._devices[device_id]  # KeyError if not found
            current_cap_traits = {t for t in self._device_traits.get(device_id, set()) if t.startswith("cap:")}
            desired_cap_traits = {f"cap:{name}" for name in device.capabilities}
            for trait in current_cap_traits - desired_cap_traits:
                self._remove_trait_index(device_id, trait)
            for trait in desired_cap_traits - current_cap_traits:
                self._add_trait_index(device_id, trait)

    def __len__(self) -> int:
        with self._lock:
            return len(self._devices)

    def __contains__(self, device_id: str) -> bool:
        with self._lock:
            return device_id in self._devices

    # ---- 內部輔助 ----

    def _add_trait_index(self, device_id: str, trait: str) -> None:
        """建立 device_id ↔ trait 的雙向索引"""
        self._device_traits[device_id].add(trait)
        if trait not in self._trait_devices:
            self._trait_devices[trait] = set()
        self._trait_devices[trait].add(device_id)

    def _remove_trait_index(self, device_id: str, trait: str) -> None:
        """移除 device_id ↔ trait 的雙向索引"""
        self._device_traits[device_id].discard(trait)
        if trait in self._trait_devices:
            self._trait_devices[trait].discard(device_id)
            # trait 已無設備時，清除該 trait 索引
            if not self._trait_devices[trait]:
                del self._trait_devices[trait]
